import collections
import itertools
import re

import pytest

from upsies.utils import types


@pytest.mark.parametrize(
    argnames='value, exp_int',
    argvalues=(
        ('1', 1), ('-1', -1),
        ('0.1', 0), ('-0.1', 0),
        ('0.9', 0), ('-0.9', 0),
        ('100', 100), (100, 100), (100.0, 100),
    ),
)
def test_Integer_valid_values(value, exp_int):
    i = types.Integer(value)
    assert i == exp_int

@pytest.mark.parametrize(
    argnames='value',
    argvalues=('zero', 'foo'),
)
def test_Integer_invalid_values(value):
    with pytest.raises(ValueError, match=rf'^Invalid integer value: {value!r}$'):
        types.Integer(value)

def test_Integer_min_max_values():
    min = 0
    max = 10

    with pytest.raises(ValueError, match=rf'^Minimum is {min}$'):
        types.Integer(min - 1, min=min, max=max)
    with pytest.raises(ValueError, match=rf'^Maximum is {max}$'):
        types.Integer(max + 1, min=min, max=max)

    for value in range(min, max + 1):
        i = types.Integer(value, min=min, max=max)
        assert i == value

    i = types.Integer(min, min=min, max=max)
    assert i.min == min
    assert i.max == max

    with pytest.raises(ValueError, match=rf'^Minimum is {min}$'):
        type(i)(min - 1)
    with pytest.raises(ValueError, match=rf'^Maximum is {max}$'):
        type(i)(max + 1)

@pytest.mark.parametrize(
    argnames='value, min, max, exp_repr',
    argvalues=(
        (5, 0, 10, 'Integer(5, min=0, max=10)'),
        (5, 0, None, 'Integer(5, min=0)'),
        (5, None, 10, 'Integer(5, max=10)'),
    ),
)
def test_Integer_repr(value, min, max, exp_repr):
    i = types.Integer(value, min=min, max=max)
    assert repr(i) == exp_repr

def test_Integer_str():
    i = types.Integer(3, min=0, max=10)
    assert str(i) == '3'


@pytest.mark.parametrize(
    argnames='value, options',
    argvalues=(
        ('1', (1, 2, 3, 4)),
        (2, (1, 2, 3, 4)),
        (3, ('1', '2', '3', '4')),
        ('4', ('1', '2', '3', '4')),
    ),
)
def test_Choice_valid_values(value, options):
    choice = types.Choice(value, options)
    assert choice == str(choice)
    assert isinstance(choice, str)

@pytest.mark.parametrize(
    argnames='empty_ok, error_expected',
    argvalues=(
        (False, True),
        (True, False),
    ),
)
def test_Choice_empty_ok(empty_ok, error_expected):
    if error_expected:
        with pytest.raises(ValueError, match=r'^Not one of bar, foo: $'):
            types.Choice('', options=('foo', 'bar'), empty_ok=empty_ok)
    else:
        choice = types.Choice('', options=('foo', 'bar'), empty_ok=empty_ok)
        assert choice == ''

@pytest.mark.parametrize(
    argnames='value, options, exp_error',
    argvalues=(
        (0, (1, 2, 3, 4), 'Not one of 1, 2, 3, 4: 0'),
        ('foo', (1, 2, 3, 4), 'Not one of 1, 2, 3, 4: foo'),
    ),
)
def test_Choice_invalid_values(value, options, exp_error):
    with pytest.raises(ValueError, match=rf'^{re.escape(exp_error)}$'):
        types.Choice(value, options)

    choice = types.Choice(options[0], options)
    with pytest.raises(ValueError, match=rf'^{re.escape(exp_error)}$'):
        type(choice)(value)

def test_Choice_repr():
    choice = types.Choice('bar', options=('foo', 'bar', 'baz'))
    assert repr(choice) == "Choice('bar', options=('bar', 'baz', 'foo'))"


@pytest.mark.parametrize(
    argnames='string, exp_bool',
    argvalues=(
        ('true', True), ('false', False),
        ('yes', True), ('no', False),
        ('1', True), ('0', False),
        ('on', True), ('off', False),
        ('aye', True), ('nay', False),
    ),
)
def test_Bool_valid_values(string, exp_bool):
    bool = types.Bool(string)
    if exp_bool:
        assert bool
    else:
        assert not bool
    assert str(bool) == string

@pytest.mark.parametrize(
    argnames='string',
    argvalues=('da', 'nyet'),
)
def test_Bool_invalid_values(string):
    with pytest.raises(ValueError, match=rf'^Invalid boolean value: {string!r}$'):
        types.Bool(string)

@pytest.mark.parametrize(
    argnames='a, b',
    argvalues=(
        (types.Bool('yes'), types.Bool('true')),
        (types.Bool('yes'), True),
        (types.Bool('yes'), 'yes'),
        (types.Bool('no'), 'no'),
    ),
)
def test_Bool_equality(a, b):
    assert a == b
    assert b == a
    assert not (a != b)
    assert not (b != a)

@pytest.mark.parametrize(
    argnames='a, b',
    argvalues=(
        (types.Bool('yes'), types.Bool('false')),
        (types.Bool('no'), types.Bool('true')),
        (types.Bool('yes'), False),
        (types.Bool('no'), True),
        (types.Bool('yes'), 'no'),
        (types.Bool('no'), 'yes'),
        (types.Bool('yes'), 0),
        (types.Bool('no'), 0),
        (types.Bool('yes'), ''),
        (types.Bool('no'), ''),
    ),
)
def test_Bool_inequality(a, b):
    assert a != b
    assert b != a
    assert not (a == b)
    assert not (b == a)

def test_Bool_attributes():
    assert types.Bool.truthy == ('true', 'yes', '1', 'on', 'aye')
    assert types.Bool.falsy == ('false', 'no', '0', 'off', 'nay')

@pytest.mark.parametrize(
    argnames='value, exp_repr',
    argvalues=(
        tuple(
            itertools.chain(
                ((value, f'Bool({value!r})') for value in types.Bool.truthy),
                ((value, f'Bool({value!r})') for value in types.Bool.falsy),
                ((value.title(), f'Bool({value.title()!r})') for value in types.Bool.truthy),
                ((value.title(), f'Bool({value.title()!r})') for value in types.Bool.falsy),
            ),
        )
    ),
)
def test_Bool___repr__(value, exp_repr):
    assert repr(types.Bool(value)) == exp_repr


@pytest.mark.parametrize('number', ('0', '1', '10', '11.5', '11.05', '99', '9999'))
@pytest.mark.parametrize('space', ('', ' '))
@pytest.mark.parametrize(
    argnames='prefix, multiplier',
    argvalues=(
        ('', 1),
        ('k', 1000),
        ('M', 1000**2),
        ('G', 1000**3),
        ('T', 1000**4),
        ('P', 1000**5),
        ('Ki', 1024),
        ('Mi', 1024**2),
        ('Gi', 1024**3),
        ('Ti', 1024**4),
        ('Pi', 1024**5),
    ),
)
@pytest.mark.parametrize('unit', ('', 'B'))
def test_Bytes_from_valid_string(number, space, prefix, multiplier, unit):
    bytes = types.Bytes(f'{number}{space}{prefix}{unit}')
    assert bytes == int(float(number) * multiplier)

@pytest.mark.parametrize(
    argnames='string, exp_msg',
    argvalues=(
        ('foo', 'Invalid size: foo'),
        ('10x', 'Invalid unit: x'),
        ('10kx', 'Invalid unit: kx'),
        ('10 Mx', 'Invalid unit: Mx'),
    ),
)
def test_Bytes_from_invalid_string(string, exp_msg):
    with pytest.raises(ValueError, match=rf'^{exp_msg}$'):
        types.Bytes(string)

@pytest.mark.parametrize(
    argnames='number, exp_string',
    argvalues=(
        (0, '0 B'),
        (1, '1 B'),
        (10, '10 B'),
        (999, '999 B'),
        (1000, '1 kB'),
        (1001, '1 kB'),
        (1000 * 1.5, '1.5 kB'),
        (1000 * 1.75, '1.75 kB'),
        (1000 * 1.799, '1.8 kB'),
        (1023, '1023 B'),
        (1024, '1 KiB'),
        (1025, '1 KiB'),
        (1024 * 1.5, '1.5 KiB'),
        (1024 * 1.75, '1.79 kB'),
        (1024 * 1.799, '1.8 KiB'),
        (1000**2, '1 MB'), (1000**3, '1 GB'), (1000**4, '1 TB'), (1000**5, '1 PB'),
        (1024**2, '1 MiB'), (1024**3, '1 GiB'), (1024**4, '1 TiB'), (1024**5, '1 PiB'),
    ),
)
def test_Bytes_as_string(number, exp_string):
    string = str(types.Bytes(number))
    assert string == exp_string

@pytest.mark.parametrize(
    argnames='number, decimal_places, prefix, trailing_zeros, exp_result',
    argvalues=(
        (0, 0, 'binary', True, '0 B'),
        (0, 0, 'decimal', False, '0 B'),
        (0, 1, 'binary', False, '0 B'),
        (0, 1, 'decimal', True, '0.0 B'),

        (1, 0, 'binary', True, '1 B'),
        (1, 0, 'decimal', False, '1 B'),
        (1, 3, 'binary', False, '1 B'),
        (1, 3, 'decimal', True, '1.000 B'),

        (999, 3, 'decimal', True, '999.000 B'),
        (1001, 3, 'decimal', True, '1.001 kB'),
        (1001, 4, 'decimal', True, '1.0010 kB'),
        (1001, 4, 'decimal', False, '1.001 kB'),

        (999, 3, 'binary', False, '999 B'),
        (1000, 2, 'binary', False, '1000 B'),
        (1000, 2, 'binary', True, '1000.00 B'),
        (1001, 1, 'binary', True, '1001.0 B'),

        (1024, 4, 'decimal', True, '1.0240 kB'),
        (1024, 3, 'decimal', True, '1.024 kB'),
        (1024, 2, 'decimal', True, '1.02 kB'),
        (1024, 1, 'decimal', True, '1.0 kB'),
        (1024, 1, 'decimal', False, '1 kB'),
        (1024, 0, 'decimal', True, '1 kB'),

        (1023, 0, 'binary', True, '1023 B'),
        (1023, 1, 'binary', True, '1023.0 B'),
        (1023, 1, 'binary', False, '1023 B'),

        (1024, 1, 'binary', True, '1.0 KiB'),
        (1024, 1, 'binary', False, '1 KiB'),
        (1024, 0, 'binary', True, '1 KiB'),

        (1025, 5, 'binary', True, '1.00098 KiB'),
        (1025, 4, 'binary', True, '1.0010 KiB'),
        (1025, 4, 'binary', False, '1.001 KiB'),
        (1025, 3, 'binary', True, '1.001 KiB'),
        (1025, 2, 'binary', True, '1.00 KiB'),
        (1025, 2, 'binary', False, '1 KiB'),

        (4000, 2, 'shortest', False, '4 kB'),
        (4096, 2, 'shortest', False, '4 KiB'),

        (100 * 1e6, 0, 'decimal', False, '100 MB'),
        (100, 3, 'foo', False, ValueError("Invalid prefix: 'foo'")),
    ),
)
def test_Bytes_format(number, prefix, decimal_places, trailing_zeros, exp_result):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            types.Bytes(number).format(
                prefix=prefix,
                decimal_places=decimal_places,
                trailing_zeros=trailing_zeros,
            )
    else:
        string = types.Bytes(number).format(
            prefix=prefix,
            decimal_places=decimal_places,
            trailing_zeros=trailing_zeros,
        )
        assert string == exp_result

def test_Bytes_repr():
    b = types.Bytes(123)
    assert repr(b) == 'Bytes(123)'


@pytest.mark.parametrize(
    argnames=('name', 'bool_value'),
    argvalues=(
        ('movie', True),
        ('series', True),
        ('season', True),
        ('episode', True),
        ('unknown', False),
    ),
)
def test_ReleaseType_truthiness(name, bool_value):
    assert bool(getattr(types.ReleaseType, name)) is bool_value

@pytest.mark.parametrize(
    argnames=('name', 'exp_str'),
    argvalues=(
        ('movie', 'movie'),
        ('season', 'season'),
        ('series', 'season'),
        ('episode', 'episode'),
        ('unknown', 'unknown'),
    ),
)
def test_ReleaseType_string(name, exp_str):
    assert str(getattr(types.ReleaseType, name)) == exp_str

@pytest.mark.parametrize(
    argnames=('name', 'exp_repr'),
    argvalues=(
        ('movie', 'ReleaseType.movie'),
        ('season', 'ReleaseType.season'),
        ('series', 'ReleaseType.season'),
        ('episode', 'ReleaseType.episode'),
        ('unknown', 'ReleaseType.unknown'),
    ),
)
def test_ReleaseType_repr(name, exp_repr):
    assert repr(getattr(types.ReleaseType, name)) == exp_repr


@pytest.mark.parametrize(
    argnames=('name', 'bool_value'),
    argvalues=(
        ('true', True),
        ('false', False),
        ('renamed', False),
        ('altered', False),
        ('unknown', False),
    ),
)
def test_SceneCheckResult_truthiness(name, bool_value):
    assert bool(getattr(types.SceneCheckResult, name)) is bool_value

@pytest.mark.parametrize(
    argnames=('name', 'exp_str'),
    argvalues=(
        ('true', 'true'),
        ('false', 'false'),
        ('renamed', 'renamed'),
        ('altered', 'altered'),
        ('unknown', 'unknown'),
    ),
)
def test_SceneCheckResult_string(name, exp_str):
    assert str(getattr(types.SceneCheckResult, name)) == exp_str

@pytest.mark.parametrize(
    argnames=('name', 'exp_repr'),
    argvalues=(
        ('true', 'SceneCheckResult.true'),
        ('false', 'SceneCheckResult.false'),
        ('renamed', 'SceneCheckResult.renamed'),
        ('altered', 'SceneCheckResult.altered'),
        ('unknown', 'SceneCheckResult.unknown'),
    ),
)
def test_SceneCheckResult_repr(name, exp_repr):
    assert repr(getattr(types.SceneCheckResult, name)) == exp_repr


@pytest.mark.parametrize(
    argnames=('pattern, exp_result'),
    argvalues=(
        (r'.*foo\.bar$', re.compile(r'.*foo\.bar$')),
        (r'*foo\.bar$', ValueError(r'*foo\.bar$: Nothing to repeat at position 0')),
    ),
)
def test_RegEx(pattern, exp_result):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            types.RegEx(pattern)
    else:
        return_value = types.RegEx(pattern)
        assert return_value == exp_result


@pytest.mark.parametrize(
    argnames=('item_type, iterable, exp_result, exp___doc__'),
    argvalues=(
        (int, ['10', '20', '45'], [10, 20, 45], 'Immutable list of int'),
        (int, None, [], 'Immutable list of int'),
        (int, ['10', 'foo', '45'], ValueError('Invalid int value: foo'), None),
        (int, 123, TypeError("'int' object is not iterable"), None),
    ),
    ids=lambda v: repr(v),
)
def test_ListOf(item_type, iterable, exp_result, exp___doc__):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            if iterable is None:
                types.ListOf(item_type)
            else:
                types.ListOf(item_type, iterable)
    else:
        if iterable is None:
            list_of = types.ListOf(item_type)
        else:
            list_of = types.ListOf(item_type, iterable)
        assert list_of == exp_result
        assert isinstance(list_of, collections.abc.Sequence)
        assert not isinstance(list_of, collections.abc.MutableSequence)

        # __doc__
        assert type(list_of).__doc__ == exp___doc__

        # __geitem__()
        for i, exp_item in enumerate(exp_result):
            assert list_of[i] == exp_item

        # __len__()
        assert len(list_of) == len(exp_result)

        # __eq__()
        assert list_of == list(exp_result)
        assert list_of == tuple(exp_result)
        unexp_result = exp_result + ['foo']
        assert list_of != list(unexp_result)
        assert list_of != tuple(unexp_result)
        assert list_of != 123
        assert list_of != 'hello'

        # __repr__()
        assert repr(list_of) == (
            type(list_of).__name__
            + '('
            + repr(tuple(exp_result))
            + ')'
        )

        # __name__
        assert type(list_of).__name__ == (
            'ListOf'
            + item_type.__name__[0].upper()
            + item_type.__name__[1:]
        )
        assert type(list_of).__qualname__ == type(list_of).__name__



import random
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies.utils import update


@pytest.mark.parametrize(
    argnames='current, release, prerelease, exp_newer_version',
    argvalues=(
        # Running stable
        ('2021.06.20', '2021.06.20', '2021.06.20', None),
        ('2021.06.20', '2021.06.20', '2021.06.21alpha', None),
        ('2021.06.20', '2021.06.20', '2021.06.21', None),
        ('2021.06.20', '2021.06.20.1', '2021.06.20', '2021.06.20.1'),
        ('2021.06.20', '2021.06.20', '2021.06.20.1alpha', None),
        ('2021.06.20.1', '2021.06.20.2', '2021.06.20', '2021.06.20.2'),
        ('2021.06.20', '2021.06.21', '2021.06.21', '2021.06.21'),
        ('2021.06.20', '2021.06.21', '2021.06.21alpha', '2021.06.21'),
        ('2021.06.20', '2021.06.21', '2021.06.22alpha', '2021.06.21'),
        ('2021.06.20', '2021.06.20', None, None),
        ('2021.06.20', '2021.06.22', None, '2021.06.22'),

        # Running prerelease
        ('2021.06.25alpha', '2021.06.20', '2021.06.20alpha', None),
        ('2021.06.25alpha', '2021.06.20', '2021.06.25alpha', None),
        ('2021.06.25alpha', '2021.06.20', '2021.06.26alpha', '2021.06.26alpha'),
        ('2021.06.25alpha', '2021.06.26', '2021.06.26alpha', '2021.06.26'),
        ('2021.06.25alpha', '2021.06.26', '2021.06.27alpha', '2021.06.27alpha'),
        ('2021.06.25alpha', '2021.06.26', '2021.06.27alpha2', '2021.06.27alpha2'),
        ('2021.06.25alpha1', '2021.06.26', '2021.06.27alpha2', '2021.06.27alpha2'),
    ),
)
@pytest.mark.asyncio
async def test_get_newer_version(current, release, prerelease, exp_newer_version, mocker):
    mocker.patch('upsies.utils.update._get_versions', AsyncMock(return_value=(current, release, prerelease)))
    newer_version = await update.get_newer_version()
    assert newer_version == exp_newer_version


@pytest.mark.parametrize(
    argnames='current, release, prerelease, exp_versions',
    argvalues=(
        ('2021.06.20', '2021.06.21', '2021.06.25alpha', ('2021.06.20', '2021.06.21', None)),
        ('2021.06.20alpha', '2021.06.21', '2021.06.25alpha', ('2021.06.20alpha', '2021.06.21', '2021.06.25alpha')),
    ),
)
@pytest.mark.asyncio
async def test_get_versions(current, release, prerelease, exp_versions, mocker):
    mocker.patch('upsies.utils.update.__version__', current)
    mocker.patch('upsies.utils.update._get_latest_release', AsyncMock(return_value=release))
    mocker.patch('upsies.utils.update._get_latest_prerelease', AsyncMock(return_value=prerelease))
    versions = await update._get_versions()
    assert versions == exp_versions


@pytest.mark.parametrize(
    argnames='versions, exp_latest_version',
    argvalues=(
        (['2021.6.20', '2023.4.7', '2021.8.10'], '2023.4.7'),
        (['2021.8.27', '2021.10.9', '2021.9.20'], '2021.10.9'),
        (['2021.10.3', '2021.10.10', '2021.10.9'], '2021.10.10'),
        ([], None),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_get_latest_release(versions, exp_latest_version, mocker):
    random.shuffle(versions)
    response_mock = Mock(
        json=Mock(return_value={
            'releases': {v: f'info about {v!r}' for v in versions},
        }),
    )
    get_mock = mocker.patch('upsies.utils.http.get', AsyncMock(return_value=response_mock))
    latest_version = await update._get_latest_release()
    if exp_latest_version:
        assert latest_version == update._fix_version(exp_latest_version)
    else:
        assert latest_version is None
    assert get_mock.call_args_list == [
        call(
            url=update._PYPI_URL,
            timeout=update._REQUEST_TIMEOUT,
            cache=True,
            max_cache_age=update._MAX_CACHE_AGE,
        ),
    ]


@pytest.mark.parametrize(
    argnames='init_file_string, exp_latest_version',
    argvalues=(
        ("foo\n__version__='2021.06.20'\nbar\n", '2021.06.20'),
        ('foo\n__version__="2021.06.20"\nbar\n', '2021.06.20'),
        ("foo\n__version__ = '2021.06.20'\nbar\n", '2021.06.20'),
        ('foo\n__version__ = "2021.06.20"\nbar\n', '2021.06.20'),
        ("foo\n__venison__='2021.06.20'\nbar\n", None),
    ),
)
@pytest.mark.asyncio
async def test_get_latest_prerelease(init_file_string, exp_latest_version, mocker):
    get_mock = mocker.patch('upsies.utils.http.get', AsyncMock(return_value=init_file_string))
    latest_version = await update._get_latest_prerelease()
    if exp_latest_version:
        assert latest_version == exp_latest_version
    else:
        assert latest_version is None
    assert get_mock.call_args_list == [
        call(
            url=update._REPO_URL,
            timeout=update._REQUEST_TIMEOUT,
            cache=True,
            max_cache_age=update._MAX_CACHE_AGE,
        ),
    ]


@pytest.mark.parametrize(
    argnames='version, exp_version',
    argvalues=(
        ('0.0.1', '0.0.1'),
        ('2021.6.20', '2021.06.20'),
        ('2021.6.20.1', '2021.06.20.1'),
        ('2021.10.20', '2021.10.20'),
    ),
)
def test_fix_version(version, exp_version):
    assert update._fix_version(version) == exp_version



"""
String formatting and parsing
"""

import importlib
import inspect
import os
import re
import sys

from .. import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


def group(name):
    """
    Return character group, e.g. "digits"

    This function simply returns an attribute from the built-in :mod:`string`
    module.

    :raise ValueError: if the built-in :mod:`string` module doesn't have
        attribute `name`
    """
    string_module = importlib.import_module('string')
    try:
        return getattr(string_module, name)
    except AttributeError:
        raise ValueError(f'No such character group: {name!r}')


_max_nfo_size = 1048576

def read_nfo(directory):
    """
    Recursively search `directory` for ``*.nfo`` files and read the first

    If `directory` is not a directory, return None.

    The nfo file is decoded with :func:`decode_nfo`.

    Files larger than 1 Mib are ignored.

    :raise ContentError: if the nfo file is not readable
    """
    error_msgs = []
    for nfo_filepath in utils.fs.file_list(directory, extensions=('nfo', 'NFO')):
        try:
            if os.path.getsize(nfo_filepath) <= _max_nfo_size:
                with open(nfo_filepath, 'rb') as f:
                    return decode_nfo(f.read())
        except OSError as e:
            error_msgs.append(e.strerror if e.strerror else str(e))

    if error_msgs:
        raise errors.ContentError(f'Failed to read nfo: {error_msgs[0]}')


def decode_nfo(bytes):
    """
    Return decoded `bytes`

    Try to decode as UTF-8 first. If that fails, decode as CP437.
    """
    try:
        return bytes.decode('utf8', 'strict')
    except UnicodeDecodeError:
        return bytes.decode('cp437', 'replace')


_capitalize_regex = re.compile(r'(\s*)(\S+)(\s*)')

def capitalize(text):
    """
    Capitalize each word in `text`

    Unlike :meth:`str.title`, only words at in front of a space or at the
    beginning of `text` are capitalized.
    """
    return ''.join(
        match.group(1) + match.group(2).capitalize() + match.group(3)
        for match in re.finditer(_capitalize_regex, text)
    )


def star_rating(rating, max_rating=10):
    """
    Return star rating string with the characters "★" (U+2605), "⯪" (U+2BEA) and
    "☆" (U+2605)

    :param float,int rating: Number between 0 and `max_rating`
    :param float,int max_rating: Maximum rating
    """
    import math
    rating = min(max_rating, max(0, rating))
    left = '\u2605' * math.floor(rating)
    if rating >= max_rating:
        middle = ''
    # Avoid floating point precision issues by rounding to 1 digit after comma
    elif round(rating % 1, 1) <= 0.3:
        middle = '\u2606'  # Empty star
    elif round(rating % 1, 1) < 0.7:
        middle = '\u2bea'  # Half star
    else:
        middle = '\u2605'  # Full star
    right = '\u2606' * (math.ceil(max_rating - rating) - 1)
    return f'{left}{middle}{right}'


if sys.version_info >= (3, 9, 0):
    def remove_prefix(string, prefix):
        return string.removeprefix(prefix)

    def remove_suffix(string, suffix):
        return string.removesuffix(suffix)

else:
    def remove_prefix(string, prefix):
        if string.startswith(prefix):
            return string[len(prefix):]
        else:
            return string

    def remove_suffix(string, suffix):
        if string.endswith(suffix):
            return string[:-len(suffix)]
        else:
            return string


def evaluate_fstring(template):
    """
    Deferred f-string evaluation

    Example:

    >>> template = 'Evaluate this: {sum((1, 2, 3)) * 2}'
    >>> evaluate_fstring(template)
    12

    References:

        https://pypi.org/project/f-yeah/
        https://stackoverflow.com/a/42497694
    """
    parent_frame = inspect.stack()[1].frame
    fstring = 'f' + repr(template)
    return eval(fstring, parent_frame.f_globals, parent_frame.f_locals)



from unittest.mock import Mock, call

import pytest

from upsies import trackers


def test_trackers(mocker):
    existing_trackers = (Mock(), Mock(), Mock())
    submodules_mock = mocker.patch('upsies.utils.submodules')
    subclasses_mock = mocker.patch('upsies.utils.subclasses', return_value=existing_trackers)
    assert trackers.trackers() == existing_trackers
    assert submodules_mock.call_args_list == [call('upsies.trackers')]
    assert subclasses_mock.call_args_list == [call(trackers.base.TrackerBase, submodules_mock.return_value)]


def test_tracker_returns_TrackerBase_instance(mocker):
    existing_trackers = (Mock(), Mock(), Mock())
    existing_trackers[0].configure_mock(name='foo')
    existing_trackers[1].configure_mock(name='bar')
    existing_trackers[2].configure_mock(name='baz')
    mocker.patch('upsies.trackers.trackers', return_value=existing_trackers)
    assert trackers.tracker('bar', x=123) is existing_trackers[1].return_value
    assert existing_trackers[1].call_args_list == [call(x=123)]

def test_tracker_fails_to_find_class(mocker):
    existing_trackers = (Mock(), Mock(), Mock())
    existing_trackers[0].configure_mock(name='foo')
    existing_trackers[1].configure_mock(name='bar')
    existing_trackers[2].configure_mock(name='baz')
    mocker.patch('upsies.trackers.trackers', return_value=existing_trackers)
    with pytest.raises(ValueError, match='^Unsupported tracker: bam$'):
        trackers.tracker('bam', x=123)
    for t in existing_trackers:
        assert t.call_args_list == []


def test_tracker_names(mocker):
    existing_trackers = (Mock(), Mock(), Mock())
    existing_trackers[0].configure_mock(name='FOO')
    existing_trackers[1].configure_mock(name='bar')
    existing_trackers[2].configure_mock(name='Baz')
    mocker.patch('upsies.trackers.trackers', return_value=existing_trackers)
    assert trackers.tracker_names() == ['bar', 'Baz', 'FOO']



"""
Hardcoded PTP metadata from upload.php
"""


ptp_tags = (
    'action',
    'adventure',
    'animation',
    'arthouse',
    'asian',
    'biography',
    'camp',
    'comedy',
    'crime',
    'cult',
    'documentary',
    'drama',
    'experimental',
    'exploitation',
    'family',
    'fantasy',
    'film.noir',
    'history',
    'horror',
    'martial.arts',
    'musical',
    'mystery',
    'performance',
    'philosophy',
    'politics',
    'romance',
    'sci.fi',
    'short',
    'silent',
    'sport',
    'thriller',
    'video.art',
    'war',
    'western',
)


ptp_subtitles = {
    # Special codes
    'No Subtitles': '44',
    'en (forced)': '50',

    # TODO: No idea what intertitles are exactly and how to detect them.
    # 'English Intertitles': '51',

    # Regular languages
    'ar': '22',     # Arabic
    'bg': '29',     # Bulgarian
    'zh': '14',     # Chinese
    'hr': '23',     # Croatian
    'cs': '30',     # Czech
    'da': '10',     # Danish
    'nl': '9',      # Dutch
    'en': '3',      # English
    'et': '38',     # Estonian
    'fi': '15',     # Finnish
    'fr': '5',      # French
    'de': '6',      # German
    'el': '26',     # Greek
    'he': '40',     # Hebrew
    'hi': '41',     # Hindi
    'hu': '24',     # Hungarian
    'is': '28',     # Icelandic
    'id': '47',     # Indonesian
    'it': '16',     # Italian
    'ja': '8',      # Japanese
    'ko': '19',     # Korean
    'lv': '37',     # Latvian
    'lt': '39',     # Lithuanian
    'no': '12',     # Norwegian
    'fa': '52',     # Persian
    'pl': '17',     # Polish
    'pt': '21',     # Portuguese
    'pt-BR': '49',  # Brazilian Port.
    'ro': '13',     # Romanian
    'ru': '7',      # Russian
    'sr': '31',     # Serbian
    'sk': '42',     # Slovak
    'sl': '43',     # Slovenian
    'es': '4',      # Spanish
    'sv': '11',     # Swedish
    'th': '20',     # Thai
    'tr': '18',     # Turkish
    'uk': '34',     # Ukrainian
    'vi': '25',     # Vietnamese
}



from unittest.mock import AsyncMock, call

import pytest

from upsies import errors
from upsies.jobs.imghost import ImageHostJob
from upsies.utils.imghosts import ImageHostBase, UploadedImage


@pytest.fixture
def imghost(tmp_path):
    class TestImageHost(ImageHostBase):
        name = 'testhost'
        default_config = {}
        upload = AsyncMock()
        _upload_image = 'not an abstract base method'
    return TestImageHost(cache_directory=tmp_path)

@pytest.fixture
async def make_ImageHostJob(tmp_path, imghost):
    def make_ImageHostJob(home_directory=tmp_path, images_total=0, enqueue=()):
        return ImageHostJob(
            home_directory=home_directory,
            cache_directory=tmp_path,
            ignore_cache=False,
            imghost=imghost,
            images_total=images_total,
            enqueue=enqueue,
        )
    return make_ImageHostJob


def test_cache_id(make_ImageHostJob):
    job = make_ImageHostJob(images_total=1)
    assert job.cache_id is None


def test_cache_directory_of_imghost(make_ImageHostJob, tmp_path):
    job = make_ImageHostJob()
    assert job._imghost.cache_directory is job.cache_directory


def test_initialize_is_called_with_enqueue_and_images_total(make_ImageHostJob):
    with pytest.raises(RuntimeError, match=(r'^You must not give both arguments '
                                            r'"enqueue" and "images_total"\.$')):
        make_ImageHostJob(images_total=24, enqueue=('a', 'b', 'c'))

def test_initialize_is_called_without_enqueue_and_images_total(make_ImageHostJob, mocker):
    job = make_ImageHostJob()
    assert job.images_total == 0


@pytest.mark.asyncio
async def test_initialize_is_called_with_enqueue(make_ImageHostJob, mocker):
    job = make_ImageHostJob(enqueue=('a', 'b', 'c'))
    assert job.images_total == 3
    job.execute()
    await job.wait()
    assert job.is_finished

@pytest.mark.asyncio
async def test_initialize_is_called_with_images_total(make_ImageHostJob, mocker):
    job = make_ImageHostJob(images_total=5)
    assert job.images_total == 5


def test_images_total_property(make_ImageHostJob):
    job = make_ImageHostJob(images_total=4)
    job.images_total = 4
    assert job.images_total == 4
    job.set_images_total(10)
    assert job.images_total == 10


@pytest.mark.asyncio
async def test_handle_input_sends_image_url(make_ImageHostJob):
    job = make_ImageHostJob(images_total=3)
    job._imghost.upload.side_effect = (
        UploadedImage('http://foo', thumbnail_url='http://foo.tiny'),
        UploadedImage('http://bar', thumbnail_url='http://bar.tiny'),
    )
    assert job.images_uploaded == 0
    assert job.urls_by_file == {}

    await job.handle_input('foo.jpg')
    assert job._imghost.upload.call_args_list == [
        call('foo.jpg', cache=not job.ignore_cache),
    ]
    assert job.output == ('http://foo',)
    assert job.uploaded_images == ('http://foo',)
    assert [i.thumbnail_url for i in job.uploaded_images] == ['http://foo.tiny']
    assert job.images_uploaded == 1
    assert job.urls_by_file == {'foo.jpg': 'http://foo'}
    assert job.urls_by_file['foo.jpg'].thumbnail_url == 'http://foo.tiny'
    assert job.errors == ()

    await job.handle_input('bar.jpg')
    assert job._imghost.upload.call_args_list == [
        call('foo.jpg', cache=not job.ignore_cache),
        call('bar.jpg', cache=not job.ignore_cache),
    ]
    assert job.output == ('http://foo', 'http://bar')
    assert [i.thumbnail_url for i in job.uploaded_images] == ['http://foo.tiny', 'http://bar.tiny']
    assert job.images_uploaded == 2
    assert job.urls_by_file == {'foo.jpg': 'http://foo', 'bar.jpg': 'http://bar'}
    assert job.urls_by_file['foo.jpg'].thumbnail_url == 'http://foo.tiny'
    assert job.urls_by_file['bar.jpg'].thumbnail_url == 'http://bar.tiny'
    assert job.errors == ()

@pytest.mark.asyncio
async def test_handle_input_handles_RequestError(make_ImageHostJob):
    job = make_ImageHostJob(images_total=3)
    job._imghost.upload.side_effect = errors.RequestError('ugly image')
    await job.handle_input('foo.jpg')
    assert job._imghost.upload.call_args_list == [
        call('foo.jpg', cache=not job.ignore_cache),
    ]
    assert job.output == ()
    assert job.errors == (errors.RequestError('ugly image'),)
    assert job.images_uploaded == 0


@pytest.mark.asyncio
async def test_exit_code(make_ImageHostJob):
    job = make_ImageHostJob(images_total=123)
    assert job.exit_code is None
    job.execute()
    assert job.exit_code is None
    job.finish()
    await job.wait()
    assert job.exit_code is not None
    job._images_uploaded = 0
    assert job.exit_code == 1
    job._images_uploaded = 122
    assert job.exit_code == 1
    job._images_uploaded = 123
    assert job.exit_code == 0



"""
Configuration files
"""

import collections
import configparser
import copy
import functools
import os
from os.path import exists as _path_exists

from .. import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


def _is_iterable_type(cls):
    return issubclass(cls, collections.abc.Iterable) and not issubclass(cls, str)

def _is_iterable(value):
    return isinstance(value, collections.abc.Iterable) and not isinstance(value, str)

def _any2iterable(value):
    return tuple(value) if _is_iterable(value) else tuple(str(value).split())

def _any2string(value):
    return ' '.join(str(v) for v in value) if _is_iterable(value) else str(value)


def config_value(value, *, description=''):
    """
    Add keyword arguments as attributes to `value`

    :param value: Any object
    :param str description: Explains what `value` is for and/or the possible
        values it can have
    """
    clsname = type(value).__name__
    bases = (type(value),)
    attrs = {
        'description': str(description),
    }
    cls = type(clsname, bases, attrs)
    return cls(value)


class ConfigFiles(collections.abc.MutableMapping):
    """
    Combine multiple INI-style configuration files into nested dictionaries

    Each top-level dictionary represents one INI file. It maps section names to
    subsections. Subsections (which are sections in the INI file) contain pairs
    of options and values.

    Sections, subsections and options can be accessed like dictionaries.

    >>> config["main"]["foo"]["bar"]
    "This is bar's value"
    >>> config["main"]["foo"]
    {'bar': "This is bar's value", 'baz': 'Another value'}
    >>> config["main"]["foo"]["bar"] = 'Zatoichi'

    For convenience, `"."` in the key is used as a delimiter.

    >>> config["main.foo.bar"] = "Ichi"
    >>> config["main.foo.bar"]
    'Ichi'

    :param defaults: Nested directory structure as described above with
        default values
    :param files: Mapping of section names to file paths that are :meth:`read`
        during initialization
    """

    def __init__(self, defaults, ignore_missing=False, **files):
        self._defaults = copy.deepcopy(defaults)
        self._cfg = _ConfigDict(self._defaults, types=self._build_types())
        self._files = {}  # Map section names to file paths
        for section, filepath in files.items():
            self.read(section, filepath, ignore_missing=ignore_missing)

    def _build_types(self):
        def make_config_type(value):
            parent_class = type(value)

            def preconvert(v):
                # Convert non-sequence to sequence if we expect a sequence
                if _is_iterable_type(parent_class):
                    if not _is_iterable(v):
                        return _any2iterable(v)
                # Convert sequence to string if we expect a string
                else:
                    if _is_iterable(v):
                        return _any2string(v)
                return v

            def __new__(cls, v):
                v = preconvert(v)
                return parent_class.__new__(cls, v)

            clsname = (
                parent_class.__name__[0].upper()
                + parent_class.__name__[1:]
                + 'ConfigType'
            )
            bases = (parent_class,)
            attrs = {
                '__new__': __new__,
            }
            return type(clsname, bases, attrs)

        types = {}
        for section, subsections in self._defaults.items():
            types[section] = {}
            for subsection, options in subsections.items():
                types[section][subsection] = {}
                for option, value in options.items():
                    types[section][subsection][option] = make_config_type(value)

        return types

    @functools.cached_property
    def paths(self):
        """Sorted tuple of ``section.subsection.option`` paths"""

        def _get_paths(dct, parents=()):
            paths = []
            for k, v in dct.items():
                k_parents = parents + (k,)
                if isinstance(v, collections.abc.Mapping):
                    paths.extend(_get_paths(v, parents=k_parents))
                else:
                    paths.append('.'.join(str(k) for k in k_parents))
            return tuple(sorted(paths))

        return _get_paths(self._defaults)

    def read(self, section, filepath, ignore_missing=False):
        """
        Read `filepath` and make its contents available as `section`

        :raises ConfigError: if reading or parsing a file fails
        """
        if ignore_missing and not _path_exists(filepath):
            self._files[section] = filepath
        else:
            try:
                with open(filepath, 'r') as f:
                    string = f.read()
            except OSError as e:
                raise errors.ConfigError(f'{filepath}: {e.strerror}')
            else:
                cfg = self._parse(section, string, filepath)
                for subsection_name, subsection in cfg.items():
                    for option_name, option_value in subsection.items():
                        try:
                            self._set(section, subsection_name, option_name, option_value)
                        except errors.ConfigError as e:
                            raise errors.ConfigError(f'{filepath}: {e}')
                self._files[section] = filepath

    def _set(self, section, subsection, option, value):
        if section not in self._cfg:
            raise errors.ConfigError(f'{section}: Unknown section')
        elif subsection not in self._cfg[section]:
            raise errors.ConfigError(f'{section}.{subsection}: Unknown subsection')
        elif option not in self._cfg[section][subsection]:
            raise errors.ConfigError(f'{section}.{subsection}.{option}: Unknown option')
        else:
            try:
                self._cfg[section][subsection][option] = value
            except errors.ConfigError as e:
                raise errors.ConfigError(f'{section}.{subsection}.{option}: {e}')

    @staticmethod
    def _parse(section, string, filepath):
        cfg = configparser.ConfigParser(
            default_section=None,
            interpolation=None,
        )
        try:
            cfg.read_string(string, source=filepath)
        except configparser.MissingSectionHeaderError as e:
            raise errors.ConfigError(f'{filepath}: Line {e.lineno}: {e.line.strip()}: Option outside of section')
        except configparser.ParsingError as e:
            lineno, msg = e.errors[0]
            raise errors.ConfigError(f'{filepath}: Line {lineno}: {msg}: Invalid syntax')
        except configparser.DuplicateSectionError as e:
            raise errors.ConfigError(f'{filepath}: Line {e.lineno}: {e.section}: Duplicate section')
        except configparser.DuplicateOptionError as e:
            raise errors.ConfigError(f'{filepath}: Line {e.lineno}: {e.option}: Duplicate option')
        except configparser.Error as e:
            raise errors.ConfigError(f'{filepath}: {e}')
        else:
            # Make normal dictionary from ConfigParser instance
            # https://stackoverflow.com/a/28990982
            cfg = {s : dict(cfg.items(s))
                   for s in cfg.sections()}

            # Line breaks are interpreted as list separators
            for section in cfg.values():
                for key in section:
                    if '\n' in section[key]:
                        section[key] = tuple(item for item in section[key].split('\n') if item)

            return cfg

    def __getitem__(self, key):
        if isinstance(key, str) and '.' in key:
            path = key.split('.') if isinstance(key, str) else list(key)
            value = self._cfg
            while path:
                value = value[path.pop(0)]
            return value
        else:
            return self._cfg[key]

    def __setitem__(self, key, value):
        if isinstance(key, str) and '.' in key:
            path = key.split('.') if isinstance(key, str) else list(key)
            target = self._cfg
            while len(path) > 1:
                key = path.pop(0)
                target = target[key]
            key = path.pop(0)
        else:
            target = self._cfg
        target[key] = value

    def __delitem__(self, key):
        self.reset(key)

    def __iter__(self):
        return iter(self._cfg)

    def __len__(self):
        return len(self._cfg)

    def __repr__(self):
        return repr(self._cfg)

    def copy(self):
        """Return deep copy as :class:`dict`"""
        return copy.deepcopy(self)

    def reset(self, path=()):
        """
        Set section, subsection or option to default value(s)

        :param path: Section, section and subsection or section, subsection and
            option as sequence or string with `"."` as delimiter
        """
        path = path.split('.') if isinstance(path, str) else list(path)
        while len(path) < 3:
            path.append('')
        self._reset(*path)

    def _reset(self, section=None, subsection=None, option=None):
        # Rename argument variables to untangle them from loop variables
        section_arg, subsection_arg, option_arg = section, subsection, option
        del section, subsection, option

        sections = (section_arg,) if section_arg else tuple(self._cfg)
        for section in sections:
            subsections = (subsection_arg,) if subsection_arg else tuple(self._cfg[section])
            for subsection in subsections:
                options = (option_arg,) if option_arg else tuple(self._cfg[section][subsection])
                for option in options:
                    self._cfg[section][subsection][option] = \
                        self._defaults[section][subsection][option]

    def write(self, *sections):
        """
        Save current configuration to file(s)

        List values use ``"\\n  "`` (newline followed by two spaces) as
        separators between items.

        :param sections: Paths to sections, subsections or options to save. Save
            all sections with no arguments.

        :raise ConfigError: if writing fails
        """
        if not sections:
            sections = tuple(self._files)
        else:
            sections = tuple(s.split('.')[0] for s in sections)

        for section in sections:
            parentdir = os.path.dirname(self._files[section])
            os.makedirs(parentdir, mode=0o700, exist_ok=True)
            try:
                with open(self._files[section], 'w') as f:
                    f.write(self._as_ini(section))
            except OSError as e:
                raise errors.ConfigError(f'{self._files[section]}: {e.strerror or e}')

    def _as_ini(self, section):
        lines = []
        for subsection, options in self._cfg.get(section, {}).items():
            lines_subsection = []
            for option, value in options.items():
                if value == self._defaults[section][subsection][option]:
                    continue

                if utils.is_sequence(value):
                    if value:
                        values = '\n  '.join(str(v) for v in value)
                        lines_subsection.append(f'{option} =\n  {values}')
                    else:
                        lines_subsection.append(f'{option} =')
                else:
                    lines_subsection.append(f'{option} = {value}')

            if lines_subsection:
                lines.append(f'[{subsection}]')
                lines.extend(lines_subsection)

                # Empty line between subsections
                lines.append('')

        return '\n'.join(lines)


class _ConfigDict(collections.abc.MutableMapping):
    """
    Dictionary that only accepts certain keys and value types

    :param dict dct: Nested mapping of initial values
    :param dict keys: Nested mapping of allowed keys. Leaf values are
        ignored. If omitted, keys from `dct` are used.
    :param dict types: Nested mapping of value converters. Converters get a
        value as a positional argument and return the converted value.

    Getting or setting keys that are not in `keys` raises `KeyError`.

    Setting keys that are dictionaries to non-dictionaries raises `TypeError`.
    Setting keys that are non-dictionaries to dictionaries raises `TypeError`.

    Subdictionaries are always instances of this class with the appropriate
    `keys` and `types`.

    Setting an existing subdictionary to a dictionary copies values instead of
    replacing the subdictionary.
    """

    def __init__(self, dct, *, keys=None, types=None):
        assert isinstance(dct, collections.abc.Mapping)
        self._keys = self._build_keys(keys or dct)
        self._types = types or {}
        self._dct = {}
        for k, v in dct.items():
            self[k] = v

    def _build_keys(self, keys):
        built_keys = {}
        for k, v in keys.items():
            if isinstance(v, collections.abc.Mapping):
                built_keys[k] = self._build_keys(v)
            else:
                built_keys[k] = None
        return built_keys

    def __getitem__(self, key):
        return self._dct[key]

    def __delitem__(self, key):
        del self._dct[key]

    def __setitem__(self, key, value):
        self._dct[key] = self._convert(key, value)

    def _convert(self, key, value):
        if key not in self._keys:
            raise KeyError(key)

        if key in self._types:
            if not isinstance(self._types[key], collections.abc.Mapping):
                converter = self._types[key]
                try:
                    value = converter(value)
                except (ValueError, TypeError) as e:
                    raise errors.ConfigError(e)

        if isinstance(self._keys[key], collections.abc.Mapping):
            # Setting a subdictionary
            if not isinstance(value, collections.abc.Mapping):
                raise TypeError(
                    f'Expected dictionary for {key}, '
                    f'not {type(value).__name__}: {value!r}'
                )
            else:
                # Merge new dictionary into existing dictionary
                if key in self._dct:
                    value = utils.merge_dicts(self._dct[key], value)
                return _ConfigDict(
                    value,
                    keys=self._keys[key],
                    types=self._types.get(key, None),
                )
        elif isinstance(value, collections.abc.Mapping):
            # Key is not a dictionary, so value can't be one
            raise TypeError(f'{key} is not a dictionary: {value!r}')
        else:
            return value

    def __iter__(self):
        return iter(self._dct)

    def __len__(self):
        return len(self._dct)

    def __repr__(self):
        return repr(self._dct)

    def copy(self):
        """Return deep copy as :class:`dict`"""
        return copy.deepcopy(self)



"""
Custom TUI widgets
"""

import collections
import itertools
import textwrap

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout.containers import HSplit, VSplit, Window, WindowAlign
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension

from ...utils import get_aioloop

import logging  # isort:skip
_log = logging.getLogger(__name__)


hspacer = Window(FormattedTextControl(' '), dont_extend_width=True, dont_extend_height=True)
vspacer = Window(FormattedTextControl(''), dont_extend_width=True, dont_extend_height=True)


class TextField:
    """Single line of non-editable text"""

    def __init__(self, text='', width=None, height=1, extend_width=True, style=''):
        self._text = text
        self._width = width
        self._height = height
        self._activity_indicator = ActivityIndicator(callback=self.set_text)
        self.container = Window(
            content=FormattedTextControl(lambda: self.text),
            width=width,
            height=height,
            dont_extend_height=True,
            dont_extend_width=not extend_width,
            wrap_lines=True,
            style=style,
        )

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text):
        self.set_text(text)

    def set_text(self, text):
        if isinstance(self._width, int) and self._width > 1:
            width = self._width
            if isinstance(self._height, int):
                height = self._height
            else:
                height = None

            if height == 1:
                # Remove newlines. Don't do word wrapping to avoid replacing
                # long words with "…" with lots of available free space.
                self._text = text.replace('\n', ' ')
            else:
                # prompt_toolkit has no support for word wrapping.
                paragraphs = text.split('\n')
                self._text = '\n\n'.join(textwrap.fill(
                    paragraph,
                    width=width,
                    max_lines=height,
                    placeholder='…',
                ) for paragraph in paragraphs)

        else:
            self._text = text

        get_app().invalidate()

    @property
    def is_loading(self):
        """Whether an activity indicator is displayed"""
        return self._activity_indicator.active

    @is_loading.setter
    def is_loading(self, is_loading):
        self._activity_indicator.active = is_loading

    def __pt_container__(self):
        return self.container


class InputField:
    """Single line of user-editable text"""

    def __init__(self, text='', width=None, extend_width=True, read_only=False,
                 on_accepted=None, on_changed=None, style=''):
        self._activity_indicator = ActivityIndicator(callback=self.set_text)
        self.read_only = read_only
        self.on_accepted = on_accepted
        self.buffer = Buffer(
            multiline=False,
            accept_handler=self._accept_handler,
            on_text_changed=on_changed,
            read_only=Condition(lambda: self.read_only),
        )
        self.container = Window(
            content=BufferControl(self.buffer),
            always_hide_cursor=Condition(lambda: self.read_only),
            width=width,
            dont_extend_height=True,
            dont_extend_width=not extend_width,
            style=style,
        )
        if text:
            self.set_text(text, ignore_callback=True)

    def _accept_handler(self, buffer):
        if self.on_accepted:
            self.on_accepted(buffer)
        # Do not clear input field on enter
        return True

    @property
    def text(self):
        return self.buffer.text

    @text.setter
    def text(self, text):
        self.set_text(text)

    def set_text(self, text, ignore_callback=False, preserve_cursor_position=False):
        if preserve_cursor_position:
            curpos = self.buffer.cursor_position

        if ignore_callback:
            handlers = self.buffer.on_text_changed._handlers
            self.buffer.on_text_changed._handlers = []

        self.buffer.set_document(Document(text), bypass_readonly=True)

        if ignore_callback:
            self.buffer.on_text_changed._handlers = handlers

        if preserve_cursor_position:
            self.buffer.cursor_position = curpos

    @property
    def read_only(self):
        return self._read_only

    @read_only.setter
    def read_only(self, read_only):
        self._read_only = bool(read_only)

    @property
    def is_loading(self):
        """Whether an activity indicator is displayed"""
        return self._activity_indicator.active

    @is_loading.setter
    def is_loading(self, is_loading):
        self._activity_indicator.active = is_loading

    def __pt_container__(self):
        return self.container


class RadioList:
    """
    List of choices the user can select from

    :param choices: Sequence of choices the user can make
    :param focused: Focused index or `None` to focus the first choice
    :param on_accepted: Callback that gets the user-picked choice

    Each choice item must be a :class:`str` or a sequence with one or more
    items. The first item is presented to the user. `on_accept` gets the
    complete choice item.
    """

    def __init__(self, choices=(), focused=None, autodetected=None, on_accepted=None):
        self.choices = choices
        self.on_accepted = on_accepted

        if focused:
            self.focused_index = choices.index(focused)
        else:
            self.focused_index = 0

        if autodetected:
            self.autodetected_index = choices.index(autodetected)
        else:
            self.autodetected_index = None

        kb = KeyBindings()

        @kb.add('up')
        @kb.add('c-p')
        @kb.add('k')
        @kb.add('s-tab')
        def _(event):
            self.focused_index = max(0, self.focused_index - 1)

        @kb.add('down')
        @kb.add('c-n')
        @kb.add('j')
        @kb.add('tab')
        def _(event):
            self.focused_index = min(len(self.choices) - 1, self.focused_index + 1)

        @kb.add('enter')
        @kb.add('c-j')
        @kb.add(' ')
        def _(event):
            if self.on_accepted is not None:
                try:
                    self.on_accepted(self.choices[self.focused_index])
                except IndexError:
                    raise RuntimeError(f'No choice at index {self.focused_index}: {self.choices}')

        self.control = FormattedTextControl(
            self._get_text_fragments,
            key_bindings=kb,
            focusable=True,
        )
        self.window = Window(
            content=self.control,
            style='class:dialog.choice',
            dont_extend_height=True,
            always_hide_cursor=True,
        )

    @property
    def focused_choice(self):
        """Currently focused item in :attr:`choices`"""
        return self.choices[self.focused_index]

    @focused_choice.setter
    def focused_choice(self, choice):
        if choice not in self.choices:
            raise ValueError(f'No such choice: {choice!r}')
        else:
            self.focused_index = self.choices.index(choice)

    def _get_text_fragments(self):
        fragments = []
        if not self.choices:
            return fragments

        def choice_as_string(choice):
            if isinstance(choice, str):
                return choice
            elif isinstance(choice, collections.abc.Sequence):
                return str(choice[0])
            else:
                raise RuntimeError(f'Invalid choice value: {choice!r}')

        width = max(len(choice_as_string(choice)) for choice in self.choices)
        for i, choice in enumerate(self.choices):
            choice_string = choice_as_string(choice)
            if i == self.focused_index:
                style = 'class:dialog.choice.focused'
                fragments.append(('[SetCursorPosition]', ''))
                fragments.append((style, ' *'))
            else:
                style = 'class:dialog.choice'
                fragments.append((style, '  '))
            fragments.append((style, ' '))
            fragments.extend(to_formatted_text(choice_string.ljust(width) + ' ', style=style))
            if i == self.autodetected_index:
                fragments.append(('class:dialog.choice', '(autodetected)'))
            fragments.append(('', '\n'))

        fragments.pop()  # Remove last newline

        return fragments

    def __pt_container__(self):
        return self.window


class HLabel:
    _groups = collections.defaultdict(lambda: [])

    def __init__(self, text, content, group=None, style=''):
        self.label = Window(
            content=FormattedTextControl(text),
            dont_extend_width=True,
            dont_extend_height=True,
            style=style,
            align=WindowAlign.RIGHT,
        )
        self.container = VSplit([
            # The HSplit makes the label non-greedy, i.e. the label itself takes
            # only one line and the space below is empty with default background
            # color.
            HSplit([self.label]),
            hspacer,
            content,
        ])

        self._group = group
        if group is not None:
            self._groups[group].append(self)
            texts = [label.label.content.text for label in self._groups[group]]
            max_width = max(len(text) for text in texts)
            for label in self._groups[group]:
                label.label.width = max_width + 1

    def __pt_container__(self):
        return self.container


class VLabel:
    def __init__(self, text, content, style=''):
        self.container = HSplit([
            Window(
                FormattedTextControl(text=text),
                dont_extend_width=False,
                dont_extend_height=True,
                style=style,
            ),
            content,
        ])

    def __pt_container__(self):
        return self.container


class ProgressBar:
    def __init__(self, text='', style=''):
        self.percent = 0
        self.container = VSplit(
            children=[
                Window(
                    style='class:info.progressbar.progress',
                    width=lambda: Dimension(weight=int(self.percent)),
                    height=1,
                ),
                Window(
                    style='class:info.progressbar',
                    width=lambda: Dimension(weight=int(100 - self.percent)),
                    height=1,
                ),
            ],
            style=style,
        )

    @property
    def percent(self):
        return self._percent

    @percent.setter
    def percent(self, value):
        self._percent = float(value)

    def __pt_container__(self):
        return self.container


class ActivityIndicator:
    """
    Activity indicator that cycles through a sequence of strings

    :param callable callback: Callable that receives a single string from
        `states` as a positional argument
    :param states: Sequence of strings that are cycled through and passed to
        `callback` individually
    :param float interval: Delay between calls to `callback`
    :param bool active: Whether `callback` is called right away
    :param str format: See :attr:`format`
    :param str style: prompt-toolkit style
    :param bool extend_width: Whether to maximize horizontally
    """

    def __init__(self, *, callback=None, states=('⠷', '⠯', '⠟', '⠻', '⠽', '⠾'),
                 interval=0.3, active=False, format='{indicator}', style='',
                 extend_width=False):
        self._iterator = itertools.cycle(states)
        self._interval = float(interval)
        self._callback = callback or None
        self._format = format
        self._text = self._format.format(indicator=states[0])
        self.active = active

        self.container = Window(
            content=FormattedTextControl(lambda: self.text),
            dont_extend_height=True,
            dont_extend_width=not extend_width,
            style=style,
        )

    @property
    def active(self):
        """Whether `callback` is being called continuously"""
        return getattr(self, '_active', False)

    @active.setter
    def active(self, active):
        if active:
            if not getattr(self, '_active', None):
                self._active = True
                self._iterate()
        else:
            self._active = False

    def enable(self):
        """Calling this method sets :attr:`active` to `True`"""
        self.active = True

    def disable(self):
        """Calling this method sets :attr:`active` to `False`"""
        self.active = False

    @property
    def format(self):
        """
        Format string

        ``{indicator}`` is replaced with the current state in `states`.
        """
        return self._format

    @format.setter
    def format(self, format):
        self._format = format

    @property
    def text(self):
        """Formatted :attr:`format` string"""
        return self._text

    def _iterate(self):
        if self.active:
            self._text = self._format.format(indicator=next(self._iterator))
            if self._callback:
                self._callback(self._text)

            get_aioloop().call_later(self._interval, self._iterate)
            get_app().invalidate()

    def __pt_container__(self):
        return self.container



"""
Base class for tracker configuration
"""

from ... import utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TrackerConfigBase(dict):
    """
    Dictionary with default values that are defined by the subclass

    Some keys defined by this base class always exist.
    """

    # Options that are available for all trackers
    _common_defaults = {
        'source': utils.configfiles.config_value(
            value='',
            description='Value of the "source" field in generated torrents.',
        ),
        'randomize_infohash': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description=(
                'Whether the info hash of generated torrents is randomized '
                'by including a random number in the metadata.\n'
                'WARNING: The tracker may choke on the non-standard field '
                'or remove the random number, forcing you to manually download '
                'the torrent.'
            ),
        ),
        'exclude': utils.configfiles.config_value(
            value=utils.types.ListOf(
                item_type=utils.types.RegEx,
                iterable=(),
            ),
            description='List of regular expressions. Matching files are excluded from generated torrents.',
        ),
        'add_to': utils.configfiles.config_value(
            value=utils.types.Choice(
                value='',
                empty_ok=True,
                options=utils.btclient.client_names(),
            ),
            description='BitTorrent client to add torrent to after submission.',
        ),
        'copy_to': utils.configfiles.config_value(
            value='',
            description='Directory path to copy torrent to after submission.',
        ),
    }

    defaults = {}
    """Default values"""

    argument_definitions = {}
    """CLI argument definitions (see :attr:`.CommandBase.argument_definitions`)"""

    def __new__(cls, config={}):
        # Merge generic and tracker-specific defaults
        combined_defaults = cls._merge(cls._common_defaults, cls.defaults)

        # Check user-given config for unknown options
        for k in config:
            if k not in combined_defaults:
                raise ValueError(f'Unknown option: {k!r}')

        # Merge user-given config with defaults
        self = super().__new__(cls)
        self.update(cls._merge(combined_defaults, config))
        return self

    @staticmethod
    def _merge(a, b):
        # Copy a
        combined = {}
        combined.update(a)

        # Update a with values from b
        for k, v in b.items():
            if k in combined:
                # Ensure same value type from a
                cls = type(combined[k])
                combined[k] = cls(v)
            else:
                # Append new value
                combined[k] = v

        return combined

    # If the config is passed as config={...}, super().__init__() will interpret
    # as a key-value pair that ends up in the config.
    def __init__(cls, *args, **kwargs):
        pass



import re

import pytest

from upsies import errors
from upsies.utils import http


@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.parametrize(
    argnames='use_tls, verify, exp_exception',
    argvalues=(
        (False, False, None),
        (True, False, None),
        (True, True, errors.RequestError('TLS: Unable to get local issuer certificate')),
        (True, False, None),
    ),
)
@pytest.mark.asyncio
async def test_request_with_tls(method, use_tls, verify, exp_exception, mock_cache, httpserver, custom_host):
    with custom_host(httpserver, 'localhost', tls=use_tls):
        httpserver.expect_request(
            uri='/',
            method=method,
        ).respond_with_data('Your response, my lord.')

        url = httpserver.url_for('/')
        if use_tls:
            assert url.startswith('https://')
        else:
            assert url.startswith('http://')

        coro = http._request(method=method, url=url, verify=verify)
        if exp_exception:
            with pytest.raises(type(exp_exception), match=rf'^{url}: {re.escape(str(exp_exception))}$'):
                await coro
        else:
            result = await coro
            assert result == 'Your response, my lord.'



"""
Timestamp parsing and normalizing
"""

def pretty(seconds):
    """
    Format `seconds` as "[[H+:]MM:]SS"

    Invalid values are returned unchanged.

    :param seconds: Number of seconds or hours, minutes and seconds as
        ":"-separated string
    :type seconds: int or float or "[[H+:]M+:]S+"

    :raise TypeError: if `seconds` is of invalid type
    :raise ValueError: if `seconds` is negative

    :return: "[[H+:]MM:]SS""
    """
    if isinstance(seconds, str):
        seconds = parse(seconds)

    if not isinstance(seconds, (int, float)):
        raise TypeError(f'Not a string or number: {seconds!r}')
    elif seconds < 0:
        raise ValueError(f'Timestamp must not be negative: {seconds!r}')

    return ':'.join((
        f'{int(seconds / 3600)}',
        f'{int(seconds % 3600 / 60):02d}',
        f'{int(seconds % 3600 % 60):02d}',
    ))


def parse(string):
    """
    Convert string format "[[H+:]MM:]SS" into integer

    :param string: Hours, minutes and seconds as ":"-separated string or number
        of seconds
    :type string: str or int or float

    :raise TypeError: if `string` is of invalid type
    :raise ValueError: if `string` has an invalid format

    :return: Number of seconds
    """
    if isinstance(string, (int, float)):
        if string >= 0:
            return string
        else:
            raise ValueError(f'Invalid timestamp: {string!r}')

    elif not isinstance(string, str):
        raise TypeError(f'Not a string or number: {string!r}')

    try:
        parts = [float(part) for part in string.split(':')]
    except ValueError:
        raise ValueError(f'Invalid timestamp: {string!r}')

    for part in parts:
        if part < 0:
            raise ValueError(f'Timestamp must not be negative: {string}')

    if len(parts) == 3:
        hours, mins, secs = parts
    elif len(parts) == 2:
        hours = 0
        mins, secs = parts
    elif len(parts) == 1:
        hours = mins = 0
        secs = parts[0]
    return (hours * 3600) + (mins * 60) + secs



from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.jobs.scene import SceneSearchJob


@pytest.fixture
async def make_SceneSearchJob(tmp_path):
    def make_SceneSearchJob(content_path=tmp_path, ignore_cache=False):
        return SceneSearchJob(
            home_directory=tmp_path,
            cache_directory=tmp_path,
            ignore_cache=ignore_cache,
            content_path=content_path,
            predb=Mock(search=AsyncMock()),
        )
    return make_SceneSearchJob


def test_cache_id(make_SceneSearchJob):
    job = make_SceneSearchJob(content_path='path/to/Foo/')
    assert job.cache_id is None


def test_initialize_with_default_predb(mocker):
    MultiPredbApi_mock = mocker.patch('upsies.utils.predbs.MultiPredbApi')
    job = SceneSearchJob(content_path='path/to/Foo/')
    assert job._predb is MultiPredbApi_mock.return_value
    assert MultiPredbApi_mock.call_args_list == [call()]

def test_initialize_with_custom_predb(mocker):
    MultiPredbApi_mock = mocker.patch('upsies.utils.predbs.MultiPredbApi')
    custom_predb = Mock()
    job = SceneSearchJob(content_path='path/to/Foo/', predb=custom_predb)
    assert job._predb is custom_predb
    assert MultiPredbApi_mock.call_args_list == []


@pytest.mark.asyncio
async def test_execute(make_SceneSearchJob, mocker):
    job = make_SceneSearchJob()
    mocker.patch.object(job, 'attach_task', Mock())
    mocker.patch.object(job, '_search', Mock())
    job.execute()
    assert job._search.call_args_list == [call()]
    assert job.attach_task.call_args_list == [call(job._search.return_value, finish_when_done=True)]


@pytest.mark.parametrize(
    argnames='exception',
    argvalues=(
        errors.SceneError('no'),
        errors.RequestError('no interwebs'),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_search_handles_exception_from_search(exception, make_SceneSearchJob, mocker):
    job = make_SceneSearchJob()
    cb = Mock()
    job.signal.register('search_results', cb)
    job._predb.search.side_effect = exception
    await job._search()
    assert job._predb.search.call_args_list == [call(
        query=job._content_path,
        only_existing_releases=False,
    )]
    assert job.output == ()
    assert job.errors == (exception,)
    assert job.exit_code == 1
    assert job.is_finished
    assert cb.call_args_list == []

@pytest.mark.asyncio
async def test_search_handles_no_results(make_SceneSearchJob, mocker):
    job = make_SceneSearchJob()
    cb = Mock()
    job.signal.register('search_results', cb)
    job._predb.search.return_value = []
    await job._search()
    assert job._predb.search.call_args_list == [call(
        query=job._content_path,
        only_existing_releases=False,
    )]
    assert job.output == ()
    assert job.errors == ('No results',)
    assert job.exit_code == 1
    assert job.is_finished
    assert cb.call_args_list == [call([])]

@pytest.mark.asyncio
async def test_search_handles_results(make_SceneSearchJob, mocker):
    job = make_SceneSearchJob()
    cb = Mock()
    job.signal.register('search_results', cb)
    job._predb.search.return_value = ['foo', 'bar', 'baz']
    await job._search()
    assert job._predb.search.call_args_list == [call(
        query=job._content_path,
        only_existing_releases=False,
    )]
    assert job.output == ('foo', 'bar', 'baz')
    assert job.errors == ()
    assert job.exit_code == 0
    assert job.is_finished
    assert cb.call_args_list == [call(['foo', 'bar', 'baz'])]



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for MTV
"""

import base64

from ...utils import argtypes, configfiles, imghosts, types
from ..base import TrackerConfigBase


class MtvTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url': base64.b64decode('aHR0cHM6Ly93d3cubW9yZXRoYW50di5tZQ==').decode('ascii'),
        'username': '',
        'password': '',
        'announce_url': configfiles.config_value(
            value='',
            description='Your personal announce URL. Automatically fetched from the website if not set.',
        ),
        'source': 'MTV',
        'image_host': types.Choice('imgbox', options=imghosts.imghost_names()),
        'screenshots': configfiles.config_value(
            value=types.Integer(4, min=3, max=10),
            description='How many screenshots to make.',
        ),
        'exclude': (
            r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$',
            r'/(?i:sample|extra|bonus|feature)',
            r'(?i:sample\.[a-z]+)$',
        ),
        'anonymous': configfiles.config_value(
            value=types.Bool('no'),
            description='Whether your username is displayed on your uploads.',
        ),
    }

    argument_definitions = {
        'submit': {
            ('--screenshots', '--ss'): {
                'help': ('How many screenshots to make '
                         f'(min={defaults["screenshots"].min}, '
                         f'max={defaults["screenshots"].max})'),
                'type': argtypes.number_of_screenshots(
                    min=defaults['screenshots'].min,
                    max=defaults['screenshots'].max,
                ),
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-title', '--ot'): {
                'help': 'Only generate title (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--ignore-dupes', '--id'): {
                'help': 'Force submission even if the tracker reports duplicates',
                'action': 'store_true',
            },
        },
    }



from unittest.mock import Mock, call

import pytest

from upsies.utils.signal import Signal


def test_keyword_arguments_are_added():
    s = Signal('foo', 'bar', 'baz')
    assert s.signals == {'foo': [], 'bar': [], 'baz': []}


def test_adding_unhashable_signal():
    s = Signal()
    with pytest.raises(TypeError, match=r'^Unhashable signal: \{\}$'):
        assert s.add({})

def test_adding_duplicate_signal():
    s = Signal('foo')
    with pytest.raises(RuntimeError, match=r"^Signal already added: 'foo'$"):
        s.add('foo')

def test_adding_adds_signal():
    s = Signal()
    assert s.signals == {}
    assert s._record_signals == []
    s.add('foo')
    assert s.signals == {'foo': []}
    assert s._record_signals == []
    s.add('bar', record=True)
    assert s.signals == {'foo': [], 'bar': []}
    assert s._record_signals == ['bar']
    s.add('baz', record=False)
    assert s.signals == {'foo': [], 'bar': [], 'baz': []}
    assert s._record_signals == ['bar']


def test_record():
    s = Signal('foo', 'bar')
    assert s.recording == []
    s.record('foo')
    assert s.recording == ['foo']
    s.record('bar')
    assert s.recording == ['foo', 'bar']
    with pytest.raises(ValueError, match=r"^Unknown signal: 'baz'$"):
        s.record('baz')
    assert s.recording == ['foo', 'bar']


def test_registering_for_unknown_signal():
    s = Signal('foo')
    with pytest.raises(ValueError, match=r"^Unknown signal: 'bar'$"):
        s.register('bar', lambda: None)

def test_registering_noncallable():
    s = Signal('foo')
    with pytest.raises(TypeError, match=r"^Not a callable: 'bar'$"):
        s.register('foo', 'bar')


@pytest.mark.parametrize('kwargs', ({}, {'1': 2}), ids=lambda v: str(v))
@pytest.mark.parametrize('args', ((), (1,), (1, 2)), ids=lambda v: str(v))
def test_emit_to_no_callback(args, kwargs):
    s = Signal('foo')
    s.emit('foo', *args, **kwargs)

@pytest.mark.parametrize('kwargs', ({}, {'1': 2}), ids=lambda v: str(v))
@pytest.mark.parametrize('args', ((), (1,), (1, 2)), ids=lambda v: str(v))
def test_emit_to_single_callback(args, kwargs):
    s = Signal('foo', 'bar')
    cb = Mock()
    s.register('foo', cb)
    s.emit('foo', *args, **kwargs)
    assert cb.call_args_list == [call(*args, **kwargs)]
    s.emit('bar', *args, **kwargs)
    assert cb.call_args_list == [call(*args, **kwargs)]

@pytest.mark.parametrize('kwargs', ({}, {'1': 2}), ids=lambda v: str(v))
@pytest.mark.parametrize('args', ((), (1,), (1, 2)), ids=lambda v: str(v))
def test_emit_to_multiple_callbacks(args, kwargs):
    s = Signal('foo', 'bar')
    cb = Mock()
    s.register('foo', cb.a)
    s.register('foo', cb.b)
    s.register('foo', cb.c)
    s.register('bar', cb.d)
    s.register('bar', cb.e)
    s.register('bar', cb.f)
    s.emit('foo', *args, **kwargs)
    assert cb.mock_calls == [
        call.a(*args, **kwargs),
        call.b(*args, **kwargs),
        call.c(*args, **kwargs),
    ]
    s.emit('bar')
    assert cb.mock_calls == [
        call.a(*args, **kwargs),
        call.b(*args, **kwargs),
        call.c(*args, **kwargs),
        call.d(),
        call.e(),
        call.f(),
    ]

def test_emit_records_call():
    s = Signal()
    s.add('foo', record=True)
    s.add('bar', record=True)
    s.add('baz', record=False)
    assert s.emissions == ()
    s.emit('foo', 123)
    s.emit('bar', 'hello', 'world')
    s.emit('foo', 'and', this='that')
    s.emit('baz', 0)
    s.emit('bar', four=56)
    assert s.emissions == (
        ('foo', {'args': (123,), 'kwargs': {}}),
        ('bar', {'args': ('hello', 'world'), 'kwargs': {}}),
        ('foo', {'args': ('and',), 'kwargs': {'this': 'that'}}),
        ('bar', {'args': (), 'kwargs': {'four': 56}}),
    )

def test_replay_emits():
    emissions = (
        ('foo', {'args': (123,), 'kwargs': {}}),
        ('bar', {'args': ('hello', 'world'), 'kwargs': {}}),
        ('foo', {'args': ('and',), 'kwargs': {'this': 'that'}}),
        ('bar', {'args': (), 'kwargs': {'four': 56}}),
    )
    s = Signal('foo', 'bar', 'baz')
    cb = Mock()
    s.register('foo', cb.foo)
    s.register('bar', cb.bar)
    s.register('baz', cb.baz)
    s.replay(emissions)
    assert cb.mock_calls == [
        call.foo(123),
        call.bar('hello', 'world'),
        call.foo('and', this='that'),
        call.bar(four=56),
    ]


def test_suspend_blocks_emissions():
    s = Signal('foo', 'bar', 'baz')
    cb = Mock()
    s.register('foo', cb.foo)
    s.register('bar', cb.bar)
    s.register('baz', cb.baz)

    s.emit('foo', 123, this='that')
    s.emit('bar', 456, hey='ho')
    s.emit('baz', 789, zip='zop')
    assert cb.mock_calls == [
        call.foo(123, this='that'),
        call.bar(456, hey='ho'),
        call.baz(789, zip='zop'),
    ]
    cb.reset_mock()

    with s.suspend('bar', 'foo'):
        s.emit('foo', 123, this='that')
        s.emit('bar', 456, hey='ho')
        s.emit('baz', 789, zip='zop')
    assert cb.mock_calls == [
        call.baz(789, zip='zop'),
    ]
    cb.reset_mock()

    s.emit('foo', 123, this='that')
    s.emit('bar', 456, hey='ho')
    s.emit('baz', 789, zip='zop')
    assert cb.mock_calls == [
        call.foo(123, this='that'),
        call.bar(456, hey='ho'),
        call.baz(789, zip='zop'),
    ]

def test_suspend_handles_unknown_signals():
    s = Signal('foo', 'bar', 'baz')
    with pytest.raises(ValueError, match=r"Unknown signal: 'boo'"):
        with s.suspend('bar', 'boo'):
            pass



"""
Generate all required metadata and upload to tracker
"""

import functools
import os

from .... import __project_name__, constants, jobs, trackers, utils
from . import base


class submit(base.CommandBase):
    """Generate all required metadata and upload to TRACKER"""

    names = ('submit',)

    argument_definitions = {}

    subcommand_name = 'TRACKER'
    subcommands = {
        tracker.name: {
            'description': (
                f'Generate all required metadata and upload it to {tracker.label}.\n'
                '\n'
                f'For step-by-step instructions run this command:\n'
                '\n'
                f'    $ {__project_name__} submit {tracker.name} --howto-setup\n'
            ),
            'cli': {
                # Default arguments for all tackers
                **{
                    ('--howto-setup',): {
                        'action': base.PrintText(text_getter=tracker.generate_setup_howto),
                        'nargs': 0,
                        'help': 'Show detailed instructions on how to do your first upload',
                    },
                    'CONTENT': {
                        'type': utils.argtypes.content,
                        'help': 'Path to release content',
                    },
                    ('--is-scene',): {
                        'type': utils.argtypes.is_scene,
                        'default': None,
                        'help': ('Whether this is a scene release (usually autodetected)\n'
                                 'Valid values: '
                                 + ', '.join(
                                     f'{true}/{false}'
                                     for true, false in zip(utils.types.Bool.truthy, utils.types.Bool.falsy)
                                 )),
                    },
                    ('--exclude-files', '--ef'): {
                        'nargs': '+',
                        'metavar': 'PATTERN',
                        'help': ('Glob pattern to exclude from torrent '
                                 '(matched case-insensitively against path in torrent)'),
                        'default': (),
                    },
                    ('--exclude-files-regex', '--efr'): {
                        'nargs': '+',
                        'metavar': 'PATTERN',
                        'help': ('Regular expression to exclude from torrent '
                                 '(matched case-sensitively against path in torrent)'),
                        'type': utils.argtypes.regex,
                        'default': (),
                    },
                    ('--reuse-torrent', '-t'): {
                        'nargs': '+',
                        'metavar': 'TORRENT',
                        'help': ('Use hashed pieces from TORRENT instead of generating '
                                 'them again or getting them from '
                                 f'{utils.fs.tildify_path(constants.GENERIC_TORRENTS_DIRPATH)}\n'
                                 'TORRENT may also be a directory, which is searched recursively '
                                 'for a matching *.torrent file.\n'
                                 "NOTE: This option is ignored if TORRENT doesn't match properly."),
                        'type': utils.argtypes.existing_path,
                        'default': (),
                    },
                    ('--add-to', '-a'): {
                        'type': utils.argtypes.client,
                        'metavar': 'CLIENT',
                        'help': ('Case-insensitive BitTorrent client name\n'
                                 'Supported clients: ' + ', '.join(utils.btclient.client_names())),
                    },
                    ('--copy-to', '-c'): {
                        'metavar': 'PATH',
                        'help': 'Copy the created torrent to PATH (file or directory)',
                    },
                },
                # Custom arguments defined by tracker for this command
                **tracker.TrackerConfig.argument_definitions.get('submit', {}),
            },
        }
        for tracker in trackers.trackers()
    }

    @functools.cached_property
    def jobs(self):
        submit_job = jobs.submit.SubmitJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            tracker=self.tracker,
            tracker_jobs=self.tracker_jobs,
        )
        return (
            tuple(self.tracker_jobs.jobs_before_upload)
            + (submit_job,)
            + tuple(self.tracker_jobs.jobs_after_upload)
        )

    @functools.cached_property
    def tracker_name(self):
        """Lower-case abbreviation of tracker name"""
        return self.args.subcommand.lower()

    @functools.cached_property
    def tracker_options(self):
        """
        :attr:`tracker_name` section in trackers configuration file combined with
        CLI arguments where CLI arguments take precedence unless their value is
        `None`
        """
        return self.get_options('trackers', self.tracker_name)

    @functools.cached_property
    def tracker(self):
        """
        :class:`~.trackers.base.TrackerBase` instance from one of the submodules of
        :mod:`.trackers`
        """
        return trackers.tracker(
            name=self.tracker_name,
            options=self.tracker_options,
        )

    @functools.cached_property
    def tracker_jobs(self):
        """
        :class:`~.trackers.base.TrackerJobsBase` instance from one of the submodules
        of :mod:`.trackers`
        """
        return self.tracker.TrackerJobs(
            options=self.tracker_options,
            content_path=self.args.CONTENT,
            reuse_torrent_path=(
                tuple(self.args.reuse_torrent)
                + tuple(self.config['config']['torrent-create']['reuse_torrent_paths'])
            ),
            tracker=self.tracker,
            screenshots_optimization=self.config['config']['screenshots']['optimize'],
            image_host=self._get_imghost(),
            btclient=self._get_btclient(),
            torrent_destination=self._get_torrent_destination(),
            exclude_files=(
                tuple(self.config['trackers'][self.tracker.name]['exclude'])
                + tuple(self.args.exclude_files)
                + tuple(self.args.exclude_files_regex)
            ),
            common_job_args={
                'home_directory': self.home_directory,
                'cache_directory': self.cache_directory,
                'ignore_cache': self.args.ignore_cache,
            },
        )

    def _get_imghost(self):
        imghost_name = self.tracker_options.get('image_host', None)
        if imghost_name:
            # Get global image host options
            options = self.config['imghosts'][imghost_name].copy()

            # Apply tracker-specific options that are common for all image hosts
            options.update(self.tracker.TrackerJobs.image_host_config.get('common', {}))

            # Apply tracker-specific options for the used image host
            options.update(self.tracker.TrackerJobs.image_host_config.get(imghost_name, {}))

            return utils.imghosts.imghost(
                name=imghost_name,
                options=options,
            )

    def _get_btclient_name(self):
        return (getattr(self.args, 'add_to', None)
                or self.tracker_options.get('add_to', None)
                or None)

    def _get_btclient(self):
        client_name = self._get_btclient_name()
        if client_name:
            options = self.get_options('clients', client_name)
            return utils.btclient.BtClient(
                name=client_name,
                url=options['url'],
                username=options['username'],
                password=options['password'],
                download_path=utils.fs.dirname(os.path.abspath(self.args.CONTENT)),
                check_after_add=options['check_after_add'],
                category=options.get('category', ''),
            )

    def _get_torrent_destination(self):
        return (getattr(self.args, 'copy_to', None)
                or self.tracker_options.get('copy_to', None)
                or None)



import configparser
import os
import stat
from unittest.mock import call, patch

import pytest

from upsies import errors
from upsies.utils import configfiles


@pytest.mark.parametrize(
    argnames=('value', 'exp_value'),
    argvalues=(
        (['a', 'b', 'c'], ('a', 'b', 'c')),
        (('a', 'b', 'c'), ('a', 'b', 'c')),
        (iter(('a', 'b', 'c')), ('a', 'b', 'c')),
        ('a b c', ('a', 'b', 'c')),
        ('abc def', ('abc', 'def')),
        (123, ('123',)),
    ),
    ids=lambda value: str(value),
)
def test_any2iterable(value, exp_value):
    assert configfiles._any2iterable(value) == exp_value

@pytest.mark.parametrize(
    argnames=('value', 'exp_value'),
    argvalues=(
        ('a b c', 'a b c'),
        (123, '123'),
        (['a', 'b', 'c'], 'a b c'),
        (('a', 'b', 'c'), 'a b c'),
        (iter(('a', 'b', 'c')), 'a b c'),
    ),
    ids=lambda value: str(value),
)
def test_any2string(value, exp_value):
    assert configfiles._any2string(value) == exp_value


@pytest.mark.parametrize(
    argnames='value, description',
    argvalues=(
        ('foo', 'This is a string'),
        (123, 'This is an int'),
        ([1, 2, 3], 'This is a list'),
    ),
)
def test_config_value(value, description):
    v = configfiles.config_value(value, description=description)
    assert v == value
    assert v.description == description


def test_ConfigFiles_init_copies_defaults():
    defaults = {'section': {'subsection1': {'foo': 123, 'bar': 456},
                            'subsection2': {'foo': 789, 'baz': 'xxx'}}}
    config = configfiles.ConfigFiles(defaults=defaults)
    assert config._defaults == defaults
    assert id(config._defaults) != id(defaults)
    assert id(config._defaults['section']) != id(defaults['section'])
    assert id(config._defaults['section']['subsection1']) != id(defaults['section']['subsection1'])
    assert id(config._defaults['section']['subsection2']) != id(defaults['section']['subsection2'])
    assert config._cfg == config._defaults
    assert isinstance(config._cfg, configfiles._ConfigDict)
    assert id(config._cfg) != id(config._defaults)

def test_ConfigFiles_init_creates_ConfigDict(mocker):
    defaults = {'section': {'subsection1': {'foo': 123, 'bar': 456},
                            'subsection2': {'foo': 789, 'baz': 'xxx'}}}
    ConfigDict_mock = mocker.patch('upsies.utils.configfiles._ConfigDict')
    mocker.patch('upsies.utils.configfiles.ConfigFiles._build_types')
    config = configfiles.ConfigFiles(defaults=defaults)
    assert config._cfg == ConfigDict_mock.return_value
    assert ConfigDict_mock.call_args_list == [
        call(defaults, types=config._build_types.return_value),
    ]

def test_ConfigFiles_init_reads_files(mocker):
    mocker.patch('upsies.utils.configfiles.ConfigFiles.read')
    config = configfiles.ConfigFiles(
        defaults={},
        foo='path/to/foo.ini',
        bar='path/to/bar.ini',
    )
    assert config.read.call_args_list == [
        call('foo', 'path/to/foo.ini', ignore_missing=False),
        call('bar', 'path/to/bar.ini', ignore_missing=False),
    ]


def test_ConfigFiles_build_types():
    class Subclass(str):
        pass

    d = configfiles.ConfigFiles(defaults={
        'section1': {
            'subsection.1': {
                'a': 'x',
            },
            'subsection.2': {
                'b': ('x',),
                (1, 2, 3): (25, 30.3),
            },
        },
        'section2': {
            'subsection.1': {
                'x': (),
            },
            'subsection_2': {
                'x': 123,
                'y': 1.23,
                'z': Subclass('hello'),
            },
        },
    })
    types = d._build_types()

    def my_assert(value, exp, type):
        assert value == exp
        assert isinstance(value, type)

    assert types['section1']['subsection.1']['a'].__name__ == 'StrConfigType'
    assert types['section1']['subsection.1']['a'].__qualname__ == 'StrConfigType'
    my_assert(types['section1']['subsection.1']['a'](1), '1', str)

    assert types['section1']['subsection.2']['b'].__name__ == 'TupleConfigType'
    assert types['section1']['subsection.2']['b'].__qualname__ == 'TupleConfigType'
    my_assert(types['section1']['subsection.2']['b'](('asdf',)), ('asdf',), tuple)
    my_assert(types['section1']['subsection.2']['b']('a s  d\nf and   \n  bar'), ('a', 's', 'd', 'f', 'and', 'bar'), tuple)
    my_assert(types['section1']['subsection.2']['b'](['a', 's', 'd\nf', 'and bar']), ('a', 's', 'd\nf', 'and bar'), tuple)

    assert types['section1']['subsection.2'][(1, 2, 3)].__name__ == 'TupleConfigType'
    assert types['section1']['subsection.2'][(1, 2, 3)].__qualname__ == 'TupleConfigType'
    my_assert(types['section1']['subsection.2'][(1, 2, 3)](123), ('123',), tuple)
    my_assert(types['section1']['subsection.2'][(1, 2, 3)]((123, 1.23)), (123, 1.23), tuple)

    assert types['section2']['subsection.1']['x'].__name__ == 'TupleConfigType'
    assert types['section2']['subsection.1']['x'].__qualname__ == 'TupleConfigType'
    my_assert(types['section2']['subsection.1']['x']([]), (), tuple)
    my_assert(types['section2']['subsection.1']['x'](('y ', ' z')), ('y ', ' z'), tuple)
    my_assert(types['section2']['subsection.1']['x']('y  z'), ('y', 'z'), tuple)

    assert types['section2']['subsection_2']['x'].__name__ == 'IntConfigType'
    assert types['section2']['subsection_2']['x'].__qualname__ == 'IntConfigType'
    my_assert(types['section2']['subsection_2']['x'](5), 5, int)
    my_assert(types['section2']['subsection_2']['x']('5'), 5, int)

    assert types['section2']['subsection_2']['y'].__name__ == 'FloatConfigType'
    assert types['section2']['subsection_2']['y'].__qualname__ == 'FloatConfigType'
    my_assert(types['section2']['subsection_2']['y'](5), 5.0, float)
    my_assert(types['section2']['subsection_2']['y'](5.3), 5.3, float)
    my_assert(types['section2']['subsection_2']['y']('5.3'), 5.3, float)

    assert types['section2']['subsection_2']['z'].__name__ == 'SubclassConfigType'
    assert types['section2']['subsection_2']['z'].__qualname__ == 'SubclassConfigType'
    my_assert(types['section2']['subsection_2']['z']('5.3'), '5.3', Subclass)
    my_assert(types['section2']['subsection_2']['z']([5, 3]), '5 3', Subclass)


def test_ConfigFiles_paths():
    d = configfiles.ConfigFiles(defaults={
        'main': {'foo': {'a': 'x'},
                 'bar': {'b': ('x',), (1, 2, 3): 'cee'}},
        0: {'foo': {'x': 'asdf'},
            None: {'y': 'fdsa', 'z': 'qux'}},
    })
    assert d.paths == tuple(sorted((
        'main.foo.a',
        'main.bar.b',
        'main.bar.(1, 2, 3)',
        '0.foo.x',
        '0.None.y',
        '0.None.z',
    )))


def test_ConfigFiles_setting_unknown_section():
    d = configfiles.ConfigFiles(defaults={
        'main': {'foo': {'a': 'x'},
                 'bar': {'b': ('x',), (1, 2, 3): 'cee'}},
        0: {'foo': {'x': 'asdf'},
            None: {'y': 'fdsa', 'z': 'qux'}},
    })
    with pytest.raises(errors.ConfigError, match=r'^nope: Unknown section$'):
        d._set('nope', 'foo', 'a', 'x')

def test_ConfigFiles_setting_unknown_subsection():
    d = configfiles.ConfigFiles(defaults={
        'main': {'foo': {'a': 'x'},
                 'bar': {'b': ('x',), (1, 2, 3): 'cee'}},
        0: {'foo': {'x': 'asdf'},
            None: {'y': 'fdsa', 'z': 'qux'}},
    })
    with pytest.raises(errors.ConfigError, match=r'^main\.nope: Unknown subsection$'):
        d._set('main', 'nope', 'a', 'x')

def test_ConfigFiles_setting_unknown_option():
    d = configfiles.ConfigFiles(defaults={
        'main': {'foo': {'a': 'x'},
                 'bar': {'b': ('x',), (1, 2, 3): 'cee'}},
        0: {'foo': {'x': 'asdf'},
            None: {'y': 'fdsa', 'z': 'qux'}},
    })
    with pytest.raises(errors.ConfigError, match=r'^main\.foo\.nope: Unknown option$'):
        d._set('main', 'foo', 'nope', 'x')

def test_ConfigFiles_setting_option_to_invalid_value():
    d = configfiles.ConfigFiles(defaults={
        'main': {'foo': {'a': 123},
                 'bar': {'b': ('x',), (1, 2, 3): 'cee'}},
        0: {'foo': {'x': 'asdf'},
            None: {'y': 'fdsa', 'z': 'qux'}},
    })
    with pytest.raises(errors.ConfigError, match=r"^main\.foo\.a: invalid literal for int\(\) with base 10: 'not a number'$"):
        d._set('main', 'foo', 'a', 'not a number')
    assert d['main.foo.a'] == 123

def test_ConfigFiles_delitem_calls_reset(mocker):
    d = configfiles.ConfigFiles(defaults={})
    mocker.patch.object(d, 'reset')
    del d['foo']
    assert d.reset.call_args_list == [call('foo')]

def test_ConfigFiles_copy(mocker):
    d1 = configfiles.ConfigFiles(defaults={
        'main': {'foo': {'a': 'x'},
                 'bar': {'b': ('x',), (1, 2, 3): (25, 30.3)}},
    })
    d2 = d1.copy()
    assert d1 == d2
    assert id(d1) != id(d2)


@patch('upsies.utils.configfiles._path_exists')
def test_ConfigFiles_read_ignores_nonexisting_path(path_exists_mock):
    path_exists_mock.return_value = False
    config = configfiles.ConfigFiles(defaults={})
    config.read('foo', 'path/to/no/such/file.ini', ignore_missing=True)
    assert config._cfg == {}
    assert config._files == {'foo': 'path/to/no/such/file.ini'}

@patch('upsies.utils.configfiles._path_exists')
def test_ConfigFiles_read_does_not_ignore_nonexisting_path(path_exists_mock):
    path_exists_mock.return_value = False
    config = configfiles.ConfigFiles(defaults={})
    with pytest.raises(errors.ConfigError, match=r'^path/to/no/such/file.ini: No such file or directory$'):
        config.read('foo', 'path/to/no/such/file.ini', ignore_missing=False)
    assert config._cfg == {}
    assert config._files == {}

@pytest.mark.parametrize('ignore_missing', (True, False), ids=lambda v: f'ignore_missing={v!r}')
def test_ConfigFiles_read_fails_to_read_file(ignore_missing, tmp_path):
    filepath = tmp_path / 'file.ini'
    filepath.write_text('abc')
    filepath.chmod(0o000)
    try:
        config = configfiles.ConfigFiles(defaults={})
        with pytest.raises(errors.ConfigError, match=rf'^{filepath}: Permission denied$'):
            config.read('foo', filepath, ignore_missing=ignore_missing)
        assert config._cfg == {}
        assert config._files == {}
    finally:
        filepath.chmod(0o660)

@pytest.mark.parametrize('ignore_missing', (True, False), ids=lambda v: f'ignore_missing={v!r}')
def test_ConfigFiles_read_updates_values(ignore_missing, tmp_path, mocker):
    defaults = {
        'main': {'foo': {'a': 'x'},
                 'bar': {'b': ('x',), 'c': 'x'}},
        'other': {'foo': {'x': 'asdf'},
                  'baz': {'y': 'fdsa', 'z': 'qux'}},
    }
    file_main = tmp_path / 'file_main.ini'
    file_main.write_text('[bar]\nc = cee\n\n[foo]\na = z\n')
    file_other = tmp_path / 'file_other.ini'
    file_other.write_text('[foo]\nx = hello\n\n[baz]\ny = world\n')

    mocker.patch('upsies.utils.configfiles.ConfigFiles._set')
    config = configfiles.ConfigFiles(defaults=defaults)

    assert config.read('main', file_main, ignore_missing=ignore_missing) is None
    assert config._set.call_args_list == [
        call('main', 'bar', 'c', 'cee'),
        call('main', 'foo', 'a', 'z'),
    ]
    assert config._files == {'main': file_main}
    assert config._defaults == defaults

    config._set.reset_mock()
    assert config.read('other', file_other, ignore_missing=ignore_missing) is None
    assert config._set.call_args_list == [
        call('other', 'foo', 'x', 'hello'),
        call('other', 'baz', 'y', 'world'),
    ]
    assert config._files == {'main': file_main, 'other': file_other}
    assert config._defaults == defaults


def test_ConfigFiles_parse_option_outside_of_section():
    with pytest.raises(errors.ConfigError, match=r"^mock/path.ini: Line 1: foo = bar: Option outside of section$"):
        configfiles.ConfigFiles._parse('section1', 'foo = bar\n', 'mock/path.ini')

def test_ConfigFiles_parse_finds_invalid_syntax():
    with pytest.raises(errors.ConfigError, match=r"^mock/path.ini: Line 2: 'foo\\n': Invalid syntax$"):
        configfiles.ConfigFiles._parse('section1', '[section]\nfoo\n', 'mock/path.ini')

def test_ConfigFiles_parse_finds_duplicate_section():
    config = configfiles.ConfigFiles(defaults={})
    with pytest.raises(errors.ConfigError, match=r'^mock/path.ini: Line 3: asdf: Duplicate section$'):
        config._parse('section1', '[asdf]\nfoo = bar\n[asdf]\na = b\n', 'mock/path.ini')

def test_ConfigFiles_parse_finds_duplicate_option():
    with pytest.raises(errors.ConfigError, match=r'^mock/path.ini: Line 3: foo: Duplicate option$'):
        configfiles.ConfigFiles._parse('section1', '[section]\nfoo = bar\nfoo = baz\n', 'mock/path.ini')

@patch('configparser.ConfigParser')
def test_ConfigFiles_parse_catches_generic_error(ConfigParser_mock):
    ConfigParser_mock.return_value.read_string.side_effect = configparser.Error('Something')
    with pytest.raises(errors.ConfigError, match=r"^mock/path.ini: Something$"):
        configfiles.ConfigFiles._parse('section1', '', 'mock/path.ini')

def test_ConfigFiles_parse_returns_nested_dictionary():
    cfg = configfiles.ConfigFiles._parse('section1', '[foo]\na = 1\nb = 2\n[bar]\n1 = one\n', 'mock/path.ini')
    assert cfg == {'foo': {'a': '1',
                           'b': '2'},
                   'bar': {'1': 'one'}}

def test_ConfigFiles_parse_uses_newlines_as_list_separator():
    cfg = configfiles.ConfigFiles._parse('section1', '[foo]\na = 1\n 2\n    3\n', 'mock/path.ini')
    assert cfg == {'foo': {'a': ('1', '2', '3')}}


def test_ConfigFiles_get_nonstring_key():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    with pytest.raises(KeyError, match=r"^\(1, 2, 3\)$"):
        config[(1, 2, 3)]

def test_ConfigFiles_get_section():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    config._set('foo', 'bar', 'qux', 456)
    config._set('hey', 'you', 'up', '!')
    section = config['foo']
    assert section == {'bar': {'baz': 'asdf', 'qux': 456}}
    assert isinstance(section, configfiles._ConfigDict)
    section = config['hey']
    assert section == {'you': {'there': 'whats', 'up': '!'}}
    assert isinstance(section, configfiles._ConfigDict)

def test_ConfigFiles_get_unknown_section():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['nope']

def test_ConfigFiles_get_subsection():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    config._set('foo', 'bar', 'qux', 456)
    config._set('hey', 'you', 'up', '!')
    subsection = config['foo']['bar']
    assert subsection == {'baz': 'asdf', 'qux': 456}
    assert isinstance(subsection, configfiles._ConfigDict)
    subsection = config['hey']['you']
    assert subsection == {'there': 'whats', 'up': '!'}
    assert isinstance(subsection, configfiles._ConfigDict)

def test_ConfigFiles_get_subsection_with_dot_notation():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    config._set('foo', 'bar', 'qux', 456)
    config._set('hey', 'you', 'up', '!')
    subsection = config['foo.bar']
    assert subsection == {'baz': 'asdf', 'qux': 456}
    assert isinstance(subsection, configfiles._ConfigDict)
    subsection = config['hey.you']
    assert subsection == {'there': 'whats', 'up': '!'}
    assert isinstance(subsection, configfiles._ConfigDict)

def test_ConfigFiles_get_unknown_subsection():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['foo']['nope']

def test_ConfigFiles_get_unknown_subsection_with_dot_notation():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['foo.nope']

def test_ConfigFiles_get_option():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    config._set('foo', 'bar', 'qux', 456)
    config._set('hey', 'you', 'up', '!')
    assert config['foo']['bar']['baz'] == 'asdf'
    assert config['foo']['bar']['qux'] == 456
    assert config['hey']['you']['there'] == 'whats'
    assert config['hey']['you']['up'] == '!'

def test_ConfigFiles_get_option_with_dot_notation():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    config._set('foo', 'bar', 'qux', 456)
    config._set('hey', 'you', 'up', '!')
    assert config['foo.bar.baz'] == 'asdf'
    assert config['foo.bar.qux'] == 456
    assert config['hey.you.there'] == 'whats'
    assert config['hey.you.up'] == '!'

def test_ConfigFiles_get_unknown_option():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['foo']['bar']['nope']

def test_ConfigFiles_get_unknown_option_with_dot_notation():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['foo.bar.nope']


def test_ConfigFiles_set_nonstring_section():
    config = configfiles.ConfigFiles(defaults={(1, 2, 3): {'bar': {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    config[(1, 2, 3)]['bar']['baz'] = 'not asdf'
    assert config[(1, 2, 3)] == {'bar': {'baz': 'not asdf', 'qux': 123}}

def test_ConfigFiles_set_nonstring_subsection():
    config = configfiles.ConfigFiles(defaults={'foo': {(1, 2, 3): {'baz': 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    config['foo'][(1, 2, 3)]['baz'] = 'not asdf'
    assert config['foo'] == {(1, 2, 3): {'baz': 'not asdf', 'qux': 123}}

def test_ConfigFiles_set_nonstring_option():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {(1, 2, 3): 'asdf', 'qux': 123}},
                                               'hey': {'you': {'there': 'whats', 'up': '?'}}})
    config['foo']['bar'][(1, 2, 3)] = 'not asdf'
    assert config['foo'] == {'bar': {(1, 2, 3): 'not asdf', 'qux': 123}}

def test_ConfigFiles_set_valid_section():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    config['foo'] = {'bar': {'baz': 'hello', 'qux': 456}}
    assert config['foo'] == {'bar': {'baz': 'hello', 'qux': 456, 'asdf': (1, 2, 3)}}

def test_ConfigFiles_set_valid_subsection():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    config['foo']['bar'] = {'baz': 'hello', 'qux': 456}
    assert config['foo'] == {'bar': {'baz': 'hello', 'qux': 456, 'asdf': (1, 2, 3)}}

def test_ConfigFiles_set_valid_option():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    config['foo']['bar']['baz'] = 'hello'
    assert config['foo'] == {'bar': {'baz': 'hello', 'qux': 123, 'asdf': (1, 2, 3)}}

def test_ConfigFiles_set_invalid_section():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    with pytest.raises(TypeError, match=r"^Expected dictionary for foo, not str: 'asdf'$"):
        config['foo'] = 'asdf'
    assert config['foo'] == {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}

def test_ConfigFiles_set_invalid_subsection():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    with pytest.raises(TypeError, match=r"^Expected dictionary for bar, not str: 'asdf'$"):
        config['foo']['bar'] = 'asdf'
    with pytest.raises(TypeError, match=r"^Expected dictionary for bar, not str: 'asdf'$"):
        config['foo.bar'] = 'asdf'
    assert config['foo'] == {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}

def test_ConfigFiles_set_converts_option_to_sequence_if_applicable():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    config['foo']['bar']['asdf'] = 'not a list'
    assert config['foo'] == {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': ('not', 'a', 'list')}}
    config['foo.bar.asdf'] = 'also not a list'
    assert config['foo'] == {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': ('also', 'not', 'a', 'list')}}

def test_ConfigFiles_set_unknown_section():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['nope'] = {'bar': {'baz': 'asdf'}}
    assert config['foo'] == {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}

def test_ConfigFiles_set_unknown_subsection():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['foo']['nope'] = {'baz': 'asdf'}
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['foo.nope'] = {'baz': 'asdf'}
    assert config['foo'] == {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}

def test_ConfigFiles_set_unknown_option():
    config = configfiles.ConfigFiles(defaults={'foo': {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}})
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['foo']['bar']['nope'] = 'hello'
    with pytest.raises(KeyError, match=r"^'nope'$"):
        config['foo.bar.nope'] = 'hello'
    assert config['foo'] == {'bar': {'baz': 'asdf', 'qux': 123, 'asdf': (1, 2, 3)}}


def test_ConfigFiles___iter__():
    config = configfiles.ConfigFiles(defaults={
        'foo': {
            'bar': {'baz': 'asdf'},
            'this': {'that': 'arf'},
        },
        'hey': {
            'you': {'there': ''},
        },
    })
    assert tuple(config) == ('foo', 'hey')


@pytest.mark.parametrize(
    argnames=('args', 'exp_args'),
    argvalues=(
        ((), ('', '', '')),
        (('foo',), ('foo', '', '')),
        (('foo.bar',), ('foo', 'bar', '')),
        (('foo.bar.baz',), ('foo', 'bar', 'baz')),
    ),
    ids=lambda value: str(value),
)
def test_ConfigFiles__reset_calls_private_reset_method(args, exp_args, mocker):
    mocker.patch('upsies.utils.configfiles.ConfigFiles._reset')
    config = configfiles.ConfigFiles(defaults={})
    config.reset(*args)
    assert config._reset.call_args_list == [call(*exp_args)]


def test_ConfigFiles__reset_everything():
    config = configfiles.ConfigFiles(defaults={
        'foo': {
            'bar': {'baz': 'asdf', 'qux': 123},
            'this': {'that': 'arf', 'qux': 456},
        },
        'hey': {
            'you': {'there': 'whats', 'up': '?'},
        },
    })
    config['foo.bar.baz'] = 'changed'
    config['foo.this.qux'] = 789
    config['hey.you.there'] = 'changed'
    config._reset()
    assert config._cfg == config._defaults

def test_ConfigFiles__reset_section():
    config = configfiles.ConfigFiles(defaults={
        'foo': {
            'bar': {'baz': 'asdf', 'qux': 123},
            'this': {'that': 'arf', 'qux': 456},
        },
        'hey': {
            'you': {'there': 'whats', 'up': '?'},
        },
    })
    config['foo.bar.baz'] = 'changed'
    config['foo.this.qux'] = 789
    config['hey.you.there'] = 'changed'
    config._reset('foo')
    assert config._cfg == {
        'foo': {
            'bar': {'baz': 'asdf', 'qux': 123},
            'this': {'that': 'arf', 'qux': 456},
        },
        'hey': {
            'you': {'there': 'changed', 'up': '?'},
        },
    }

def test_ConfigFiles__reset_subsection():
    config = configfiles.ConfigFiles(defaults={
        'foo': {
            'bar': {'baz': 'asdf', 'qux': 123},
            'this': {'that': 'arf', 'qux': 456},
        },
        'hey': {
            'you': {'there': 'whats', 'up': '?'},
        },
    })
    config['foo.bar.baz'] = 'changed'
    config['foo.this.qux'] = 789
    config['hey.you.there'] = 'changed'
    config._reset('foo', 'this')
    assert config._cfg == {
        'foo': {
            'bar': {'baz': 'changed', 'qux': 123},
            'this': {'that': 'arf', 'qux': 456},
        },
        'hey': {
            'you': {'there': 'changed', 'up': '?'},
        },
    }

def test_ConfigFiles__reset_option():
    config = configfiles.ConfigFiles(defaults={
        'foo': {
            'bar': {'baz': 'asdf', 'qux': 123},
            'this': {'that': 'arf', 'qux': 456},
        },
        'hey': {
            'you': {'there': 'whats', 'up': '?'},
        },
    })
    config['foo.bar.baz'] = 'changed'
    config['foo.this.that'] = 'arf and arf'
    config['foo.this.qux'] = 789
    config['hey.you.there'] = 'changed'
    config._reset('foo', 'this', 'qux')
    assert config._cfg == {
        'foo': {
            'bar': {'baz': 'changed', 'qux': 123},
            'this': {'that': 'arf and arf', 'qux': 456},
        },
        'hey': {
            'you': {'there': 'changed', 'up': '?'},
        },
    }


@pytest.mark.parametrize(
    argnames='args',
    argvalues=(
        (),
        ('File_1',),
        ('File_2',),
        ('File_1', 'File_2'),
        ('File_1.foo', 'File_2.foo.bar'),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.parametrize('files_exist', (False, True))
def test_ConfigFiles_write_writes_files(files_exist, args, tmp_path):
    file1 = tmp_path / 'subdir' / 'subsubdir' / 'file1.ini'
    file2 = tmp_path / 'subdir' / 'subsubdir' / 'file2.ini'
    if files_exist:
        file1.parent.mkdir(parents=True, exist_ok=True)
        file1.write_text('[foo]\nbar = baz\n')
        file2.parent.mkdir(parents=True, exist_ok=True)
        file2.write_text('[foo]\nbar = baz\n')
    config = configfiles.ConfigFiles(
        defaults={
            'File_1': {'foo': {'bar': 'qux'}},
            'File_2': {'foo': {'bar': 'qux'}},
        },
        File_1=file1,
        File_2=file2,
        ignore_missing=True,
    )
    config['File_1.foo.bar'] = 'config written'
    config['File_2.foo.bar'] = 'config written'

    # Make sure umask isn't set so we get expected permissions by default
    umask_orig = os.umask(0o022)
    try:
        config.write(*args)
    finally:
        os.umask(umask_orig)

    # Unless files pre-existed, permissions should be owner-only
    if not files_exist:
        for f in (file1, file2):
            parent = f.parent
            assert stat.filemode(parent.stat().st_mode) == 'drwx------'

    if not args:
        files_changed = (file1, file2)
        files_unchanged = ()
    else:
        files_changed = []
        files_unchanged = []
        if any(arg.startswith('File_1') for arg in args):
            files_changed.append(file1)
        else:
            files_unchanged.append(file1)
        if any(arg.startswith('File_2') for arg in args):
            files_changed.append(file2)
        else:
            files_unchanged.append(file2)
    for f in files_changed:
        assert open(f).read() == '[foo]\nbar = config written\n'
    for f in files_unchanged:
        if files_exist:
            assert open(f).read() != '[foo]\nbar = config written\n'
        else:
            assert not os.path.exists(f)

@pytest.mark.parametrize(argnames='arg', argvalues=('main', 'main.foo'))
def test_ConfigFiles_write_extracts_section_from_argument(arg, tmp_path):
    file = tmp_path / 'file1.ini'
    file.write_text('[foo]\nbar = baz\n')
    config = configfiles.ConfigFiles(
        defaults={
            'main': {'foo': {'bar': 'qux'}},
        },
        main=file,
    )
    config['main.foo.bar'] = 'config written'
    config.write(arg)
    assert open(file).read() == '[foo]\nbar = config written\n'

def test_ConfigFiles_write_fails_to_write_file(tmp_path):
    file = tmp_path / 'file1.ini'
    file.write_text('[foo]\nbar = baz\n')
    config = configfiles.ConfigFiles(
        defaults={'main': {'foo': {'bar': 'qux'}}},
        main=file,
    )
    os.chmod(file, 0o000)
    try:
        with pytest.raises(errors.ConfigError, match=rf'^{file}: Permission denied$'):
            config.write('main')
    finally:
        os.chmod(file, 0o600)


def test_ConfigFiles_as_ini():
    config = configfiles.ConfigFiles(
        defaults={'main': {
            'foo': {'a': 'aaa', 'b': [1, 2, 3], 'c': 123},
            'bar': {'b': 'bbb', 'c': [4, 5, 6], 'd': 456},
            'baz': {'c': 'ccc', 'd': [7, 8, 9], 'e': 789},
        }},
    )
    config['main.foo.a'] = 'AAA'
    config['main.bar.c'] = [100, 200, 300]
    config['main.baz.d'] = []
    config['main.baz.e'] = 1000
    exp_ini_string = (
        '[foo]\n'
        'a = AAA\n'
        '\n'
        '[bar]\n'
        'c =\n'
        '  100\n'
        '  200\n'
        '  300\n'
        '\n'
        '[baz]\n'
        'd =\n'
        'e = 1000\n'
    )
    assert config._as_ini('main') == exp_ini_string


def test_ConfigDict_implements_delitem():
    d = configfiles._ConfigDict({'foo': 'bar', 'bar': 'abc'})
    del d['bar']
    assert d == {'foo': 'bar'}

def test_ConfigDict_implements_len():
    d = configfiles._ConfigDict({'foo': 'baz', 'bar': 'xyz'})
    assert len(d) == 2

def test_ConfigDict_implements_repr():
    d = configfiles._ConfigDict({'foo': 'baz', 'bar': 'xyz'})
    assert repr(d) == repr({'foo': 'baz', 'bar': 'xyz'})

def test_ConfigDict_implements_getitem():
    d = configfiles._ConfigDict({'foo': 'bar', 'baz': 123})
    assert d['foo'] == 'bar'
    assert d['baz'] == 123

def test_ConfigDict_implements_setitem():
    d = configfiles._ConfigDict({'foo': 'bar', 'baz': 123})
    d['foo'] = 'hello'
    d['baz'] = 'world'
    assert d == {'foo': 'hello', 'baz': 'world'}

def test_ConfigDict_raises_KeyError_when_setting_unknown_key():
    d = configfiles._ConfigDict({'foo': 'bar'})
    with pytest.raises(KeyError, match=r"^'asdf'$"):
        d['asdf'] = 'hello'

def test_ConfigDict_with_custom_keys():
    d = configfiles._ConfigDict({'foo': 'bar'}, keys={'foo': None, 'bar': None, 123: None})
    d['foo'] = 'asdf'
    d['bar'] = 'baz'
    d[123] = '456'
    assert d == {'foo': 'asdf', 'bar': 'baz', 123: '456'}
    with pytest.raises(KeyError, match=r"^'asdf'$"):
        d['asdf'] = 'hello'

@pytest.mark.parametrize(
    argnames=('converter', 'value', 'exp_value'),
    argvalues=(
        (int, 13.5, 13),
        (str, 13.5, '13.5'),
        (lambda value: str(value).split(','), 'a,b,c', ['a', 'b', 'c']),
        (lambda value: [int(v) for v in str(value).split(',')], '1,2,3', [1, 2, 3]),
    ),
    ids=lambda value: str(value),
)
def test_ConfigDict_with_custom_types(converter, value, exp_value):
    d = configfiles._ConfigDict(dct={'foo': 2, 'bar': ''}, types={'foo': converter})
    d['foo'] = value
    assert d['foo'] == exp_value
    for v in ('abc', 123, (1, 2, 3)):
        d['bar'] = v
        assert d['bar'] is v

def test_ConfigDict_converter_raises_ValueError():
    dct = {0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}}
    d = configfiles._ConfigDict(dct=dct, types={0: {1: int}})
    with pytest.raises(errors.ConfigError, match=r"^invalid literal for int\(\) with base 10: 'hello'$"):
        d[0][1] = 'hello'

def test_ConfigDict_converter_raises_TypeError():
    dct = {0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}}
    d = configfiles._ConfigDict(dct=dct, types={0: {1: int}})
    with pytest.raises(errors.ConfigError, match=(r"^int\(\) argument must be a string, "
                                                  r"a bytes-like object or a (?:real |)number, not 'list'$")):
        d[0][1] = ['hello']

def test_ConfigDict_subdictionaries_are_ConfigDicts():
    d = configfiles._ConfigDict({0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}})
    assert isinstance(d[0], configfiles._ConfigDict)
    assert isinstance(d[0][3], configfiles._ConfigDict)
    assert isinstance(d[0][3][8], configfiles._ConfigDict)

def test_ConfigDict_setting_subdictionary_merges_into_current_values():
    dct = {0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}}
    d = configfiles._ConfigDict(dct=dct)
    d[0][3] = {6: 600, 8: {9: 1000}}
    assert d == {0: {1: 2, 3: {4: 5, 6: 600, 8: {9: 1000}}}}
    assert isinstance(d[0], configfiles._ConfigDict)
    assert isinstance(d[0][3], configfiles._ConfigDict)
    assert isinstance(d[0][3][8], configfiles._ConfigDict)

def test_ConfigDict_setting_subdictionary_creates_new_subdictionary():
    dct = {0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}}
    d = configfiles._ConfigDict(dct=dct)
    del d[0][3]
    d[0][3] = {8: {9: 1000}}
    assert d == {0: {1: 2, 3: {8: {9: 1000}}}}
    assert isinstance(d[0], configfiles._ConfigDict)
    assert isinstance(d[0][3], configfiles._ConfigDict)
    assert isinstance(d[0][3][8], configfiles._ConfigDict)

def test_ConfigDict_setting_subdictionaries_copies_appropriate_subkeys():
    d = configfiles._ConfigDict({0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}},
                                keys={0: {1: None, 3: {4: None, 6: None, 8: {9: None}}}})
    assert d[0]._keys == {1: None, 3: {4: None, 6: None, 8: {9: None}}}
    assert d[0][3]._keys == {4: None, 6: None, 8: {9: None}}
    assert d[0][3][8]._keys == {9: None}

def test_ConfigDict_setting_subdictionaries_copies_appropriate_subtypes():
    types = {0: {1: lambda x: None, 3: {4: lambda x: None, 6: lambda x: None, 8: {9: lambda x: None}}}}
    d = configfiles._ConfigDict({0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}}, types=types)
    assert d[0]._types is types[0]
    assert d[0][3]._types is types[0][3]
    assert d[0][3][8]._types is types[0][3][8]

def test_ConfigDict_setting_subdictionary_to_nondictionary():
    d = configfiles._ConfigDict({0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}})
    with pytest.raises(TypeError, match=r'^Expected dictionary for 8, not int: 100$'):
        d[0][3][8] = 100

def test_ConfigDict_nondictionary_value_to_dictionary():
    d = configfiles._ConfigDict({0: {1: 2, 3: {4: 5, 6: 7, 8: {9: 0}}}})
    with pytest.raises(TypeError, match=r'^4 is not a dictionary: \{400: 4000\}$'):
        d[0][3][4] = {400: 4000}

def test_ConfigDict_copy():
    d1 = configfiles._ConfigDict({0: {1: 2, 3: {4: 5, 6: 7, 8: {9: ['a', 'b', 'c']}}}})
    d2 = d1.copy()
    assert d2 == {0: {1: 2, 3: {4: 5, 6: 7, 8: {9: ['a', 'b', 'c']}}}}
    assert id(d2) != id(d1)
    assert id(d2[0][3][8][9]) != id(d1[0][3][8][9])



import asyncio
import re
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors, utils


def test_submodules_finds_modules_and_packages(mocker):
    mocker.patch('os.listdir', return_value=(
        '_this_is_private',
        'this_is_a_package',
        'this_is_a_module.py',
        'this_is_something_else.foo',
    ))
    import_mock = mocker.patch('importlib.import_module', side_effect=(1, 2, 3, 4, 5))
    assert set(utils.submodules('asdf')) == {1, 2}
    assert import_mock.call_args_list == [
        call(name='.this_is_a_package', package='asdf'),
        call(name='.this_is_a_module', package='asdf'),
    ]


def test_subclasses():
    from aiobtclientapi.clients import APIBase, deluge, qbittorrent, rtorrent, transmission
    subclses = utils.subclasses(
        basecls=APIBase,
        modules={deluge, qbittorrent, rtorrent, transmission},
    )
    assert subclses == {
        deluge.DelugeAPI,
        qbittorrent.QbittorrentAPI,
        rtorrent.RtorrentAPI,
        transmission.TransmissionAPI,
    }


def test_closest_number():
    assert utils.closest_number(5, ()) == 0
    assert utils.closest_number(5, (), default=123) == 123

    numbers = (10, 20, 30)

    for n in (-10, 0, 9, 10, 11, 14):
        assert utils.closest_number(n, numbers) == 10
    for n in (16, 19, 20, 21, 24):
        assert utils.closest_number(n, numbers) == 20
    for n in range(26, 50):
        assert utils.closest_number(n, numbers) == 30

    for n in range(0, 50):
        with pytest.raises(ValueError, match=r'^No number equal to or below 9: \(10, 20, 30\)$'):
            utils.closest_number(n, numbers, max=9)

    for n in range(0, 50):
        assert utils.closest_number(n, numbers, max=10) == 10

    for n in range(0, 16):
        assert utils.closest_number(n, numbers, max=20) == 10
    for n in range(16, 50):
        assert utils.closest_number(n, numbers, max=20) == 20


def test_CaseInsensitiveString_equality():
    assert utils.CaseInsensitiveString('Foo') == 'foo'
    assert utils.CaseInsensitiveString('fOo') == 'FOO'
    assert utils.CaseInsensitiveString('foO') == 'foo'
    assert utils.CaseInsensitiveString('foo') != 'fooo'

def test_CaseInsensitiveString_identity():
    assert utils.CaseInsensitiveString('Foo') in ('foo', 'bar', 'baz')
    assert utils.CaseInsensitiveString('fOo') in ('FOO', 'BAR', 'BAZ')
    assert utils.CaseInsensitiveString('foO') not in ('fooo', 'bar', 'baz')

def test_CaseInsensitiveString_lt():
    assert utils.CaseInsensitiveString('foo') < 'Fooo'

def test_CaseInsensitiveString_le():
    assert utils.CaseInsensitiveString('foo') <= 'Foo'

def test_CaseInsensitiveString_gt():
    assert utils.CaseInsensitiveString('Fooo') > 'foo'

def test_CaseInsensitiveString_ge():
    assert utils.CaseInsensitiveString('Foo') >= 'foo'


def test_MonitoredList_getitem():
    lst = utils.MonitoredList((1, 2, 3), callback=Mock())
    assert lst[0] == 1
    assert lst[1] == 2
    assert lst[2] == 3

def test_MonitoredList_setitem():
    cb = Mock()
    lst = utils.MonitoredList((1, 2, 3), callback=cb)
    assert cb.call_args_list == []
    lst[0] = 100
    assert cb.call_args_list == [call(lst)]
    assert lst == [100, 2, 3]

def test_MonitoredList_delitem():
    cb = Mock()
    lst = utils.MonitoredList((1, 2, 3), callback=cb)
    assert cb.call_args_list == []
    del lst[1]
    assert cb.call_args_list == [call(lst)]
    assert lst == [1, 3]

def test_MonitoredList_insert():
    cb = Mock()
    lst = utils.MonitoredList((1, 2, 3), callback=cb)
    assert cb.call_args_list == []
    lst.insert(1, 1.5)
    assert cb.call_args_list == [call(lst)]
    assert lst == [1, 1.5, 2, 3]

def test_MonitoredList_len():
    lst = utils.MonitoredList((1, 2, 3), callback=Mock())
    assert len(lst) == 3

def test_MonitoredList_equality():
    lst = utils.MonitoredList((1, 2, 3), callback=Mock())
    assert lst == [1, 2, 3]
    assert lst != [1, 2, 4]
    assert lst == utils.MonitoredList([1, 2, 3], callback=Mock())
    assert lst != utils.MonitoredList([1, 3, 2], callback=Mock())

def test_MonitoredList_repr():
    cb = Mock()
    lst = utils.MonitoredList((1, 2, 3), callback=cb)
    assert repr(lst) == f'MonitoredList([1, 2, 3], callback={cb!r})'


@pytest.mark.parametrize(
    argnames=('args, kwargs, exp_mapping'),
    argvalues=(
        (({'foo': 'bar'},), {}, {'foo': 'bar'}),
        (({('f', 'o', 'o'): 'bar'},), {}, {('f', 'o', 'o'): 'bar'}),
        (({'foo': ['b', 'a', 'r']},), {}, {'foo': ('b', 'a', 'r')}),
        (({'foo': ['b', 'a', 'r', {1: 2}]},), {}, {'foo': ('b', 'a', 'r', frozenset(((1, 2))))}),
        (({'foo': {1: 'b', 2: ['a', 'r']}},), {}, {'foo': frozenset(((1, 'b'), (2, ('a', 'r'))))}),
    ),
    ids=lambda v: str(v),
)
def test_ImmutableMapping(args, kwargs, exp_mapping):
    mapping = utils.ImmutableMapping(*args, **kwargs)
    assert mapping == exp_mapping


def test_is_sequence():
    assert utils.is_sequence((1, 2, 3))
    assert utils.is_sequence([1, 2, 3])
    assert not utils.is_sequence('123')


@pytest.mark.parametrize(
    argnames=('a', 'b', 'merged'),
    argvalues=(
        ({1: 10}, {1: 20}, {1: 20}),
        ({1: 10}, {2: 20}, {1: 10, 2: 20}),
        ({1: 10}, {1: 20, 2: 2000}, {1: 20, 2: 2000}),
        ({1: 10, 2: 2000}, {1: 20}, {1: 20, 2: 2000}),
        ({1: {2: 20}}, {1: {2: 2000}}, {1: {2: 2000}}),
        ({1: {2: 20}}, {1: {2: {3: 30}}}, {1: {2: {3: 30}}}),
        ({1: {2: 20, 3: {5: 50}}}, {1: {3: {4: 40, 5: 5000}}}, {1: {2: 20, 3: {4: 40, 5: 5000}}}),
    ),
    ids=lambda v: str(v),
)
def test_merge_dicts(a, b, merged):
    assert utils.merge_dicts(a, b) == merged
    assert id(a) != id(merged)
    assert id(b) != id(merged)


def test_deduplicate_deduplicates():
    assert utils.deduplicate([1, 2, 1, 3, 1, 4, 3, 5]) == [1, 2, 3, 4, 5]

def test_deduplicate_maintains_order():
    assert utils.deduplicate([3, 2, 1, 1, 4, 5, 1, 3]) == [3, 2, 1, 4, 5]

def test_deduplicate_deduplicates_unhashable_items():
    items = [
        {'a': 1, 'b': 2},
        {'a': 2, 'b': 2},
        {'a': 3, 'b': 1},
        {'a': 4, 'b': 3},
        {'a': 5, 'b': 1},
        {'a': 6, 'b': 4},
    ]
    assert utils.deduplicate(items, key=lambda item: item['b']) == [
        {'a': 1, 'b': 2},
        {'a': 3, 'b': 1},
        {'a': 4, 'b': 3},
        {'a': 6, 'b': 4},
    ]


@pytest.mark.parametrize(
    argnames=('items', 'group_sizes', 'exp_groups'),
    argvalues=(
        (range(18), [1], [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,), (11,), (12,), (13,), (14,), (15,), (16,), (17,)]),
        (range(18), [1, 2], [(0, 1), (2, 3), (4, 5), (6, 7), (8, 9), (10, 11), (12, 13), (14, 15), (16, 17)]),
        (range(18), [1, 2, 3], [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9, 10, 11), (12, 13, 14), (15, 16, 17)]),
        (range(18), [1, 2, 3, 4], [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9, 10, 11), (12, 13, 14), (15, 16, 17)]),
        (range(18), [1, 2, 3, 4, 5, 6], [(0, 1, 2, 3, 4, 5), (6, 7, 8, 9, 10, 11), (12, 13, 14, 15, 16, 17)]),

        (range(17), [1], [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,), (11,), (12,), (13,), (14,), (15,), (16,)]),
        (range(17), [1, 2], [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,), (11,), (12,), (13,), (14,), (15,), (16,)]),
        (range(17), [1, 2, 3], [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,), (11,), (12,), (13,), (14,), (15,), (16,)]),
        (range(17), [2, 3, 4], [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9, 10, 11), (12, 13, 14), (15, 16, -1)]),
        (range(17), [3, 4], [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9, 10, 11), (12, 13, 14), (15, 16, -1)]),
        (range(17), [4], [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15), (16, -1, -1, -1)]),
        (range(17), [4, 5], [(0, 1, 2, 3, 4), (5, 6, 7, 8, 9), (10, 11, 12, 13, 14), (15, 16, -1, -1, -1)]),

        (range(16), [4], [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15)]),
        (range(16), [5, 4], [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15)]),
        (range(16), [4, 5, 3], [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9, 10, 11), (12, 13, 14, 15)]),
    ),
)
def test_as_groups(items, group_sizes, exp_groups):
    groups = tuple(utils.as_groups(items, group_sizes=group_sizes, default=-1))
    assert len(groups) == len(exp_groups)
    for a, b in zip(groups, exp_groups):
        assert a == b


@pytest.mark.parametrize(
    argnames='object, exp_return_value',
    argvalues=(
        ('foo', False),
        ((1, 2, 3), False),
        (re.compile(r'1, 2, 3'), True),
    ),
)
def test_is_regex_pattern(object, exp_return_value, mocker):
    assert utils.is_regex_pattern(object) is exp_return_value


@pytest.mark.parametrize(
    argnames='obj, exp_id',
    argvalues=(
        # Strings and other simple objects
        ('foo', b'foo'),
        (23, b'23'),
        (2.3, b'2.3'),

        # Iterables
        (['foo', 'bar', 'baz', 4], b'4barbazfoo'),
        (['foo', 'bar', 'baz', (4, 'five')], b'4fivebarbazfoo'),
        ([('foo', 'bar', 'baz'), {'hey': 'ho'}], b'barbazfooheyho'),
        ([('foo', 'bar', 'baz'), {'hey': ['h', 0]}], b'0hheybarbazfoo'),
        ([(1, 2, 3), ('5', 6.0), {7: ['eigh', 't']}], b'12356.07eight'),
        ([(1, 2, 3, (4.1, 4.2)), ('5', 6.0), {7: ['eigh', 't']}], b'1234.14.256.07eight'),

        # Mappings
        ({'a': 1, 'b': 2, 'c': 3}, b'1a2b3c'),
        ({('a', 'b'): 1, 'c': 2}, b'1ab2c'),
        ({('a', 'b', (1, 2)): 'what', 'c': 3}, b'12abwhat3c'),
        ({('a', 'b', (1, (2, 3))): 'what', 'c': 3}, b'123abwhat3c'),
        ({'what': ('a', 'b', (1, (2, 3))), 'c': 3}, b'123abwhat3c'),
    ),
    ids=lambda v: str(v),
)
def test_semantic_hash(obj, exp_id, mocker):
    sha256_mock = mocker.patch('hashlib.sha256')
    assert utils.semantic_hash(obj) == sha256_mock.return_value.hexdigest.return_value
    assert sha256_mock.call_args_list == [call(exp_id)]

@pytest.mark.parametrize(
    argnames='obj',
    argvalues=(
        (i for i in range(10)),
    ),
    ids=lambda v: repr(v),
)
def test_semantic_hash_gets_unsupported_type(obj, mocker):
    sha256_mock = mocker.patch('hashlib.sha256')
    with pytest.raises(RuntimeError, match=rf'Unsupported type: {type(obj)!r}: {obj!r}'):
        utils.semantic_hash(obj)
    assert sha256_mock.call_args_list == []


@pytest.fixture
def unhandled_exception_handler(event_loop):
    original_exception_handler = event_loop.get_exception_handler()
    unhandled_exception_handler = Mock()
    event_loop.set_exception_handler(unhandled_exception_handler)
    yield unhandled_exception_handler
    event_loop.set_exception_handler(original_exception_handler)

def assert_unhandled_exception(handler, exception):
    if not exception:
        assert handler.call_args_list == []
    else:
        print(handler.call_args_list)
        assert len(handler.call_args_list) == 1
        assert len(handler.call_args_list[0].args) == 2
        assert len(handler.call_args_list[0].kwargs) == 0
        args = handler.call_args_list[0].args
        assert isinstance(args[0], asyncio.AbstractEventLoop)
        assert isinstance(args[1], dict)
        raised_exception = args[1].get('exception', 'NO EXCEPTION')
        assert raised_exception is exception

@pytest.mark.parametrize('callback', (None, Mock()))
@pytest.mark.asyncio
async def test_run_task__returned_task_returns_coro_return_value(callback, unhandled_exception_handler):
    coro = AsyncMock(return_value='foo')()
    task = utils.run_task(coro, callback=callback)

    return_value = await task
    assert return_value == 'foo'

    if callback:
        assert callback.call_args_list == [call(return_value)]

    assert_unhandled_exception(unhandled_exception_handler, None)

@pytest.mark.parametrize('callback', (None, Mock()))
@pytest.mark.asyncio
async def test_run_task__coro_raises_CancelledError(callback, unhandled_exception_handler):
    coro = AsyncMock(side_effect=asyncio.CancelledError())()
    task = utils.run_task(coro, callback=callback)

    return_value = await task
    assert return_value is None

    if callback:
        assert callback.call_args_list == [call(return_value)]

    assert_unhandled_exception(unhandled_exception_handler, None)

@pytest.mark.parametrize('callback', (None, Mock()))
@pytest.mark.asyncio
async def test_run_task__coro_raises_Exception(callback, unhandled_exception_handler):
    exception = errors.UpsiesError('my exception')
    coro = AsyncMock(side_effect=exception)()

    task = utils.run_task(coro, callback=callback)
    with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
        await task

    if callback:
        assert callback.call_args_list == [call(errors.UpsiesError('my exception'))]
        assert_unhandled_exception(unhandled_exception_handler, None)
    else:
        assert_unhandled_exception(unhandled_exception_handler, exception)

@pytest.mark.asyncio
async def test_run_task__callback_raises_Exception(unhandled_exception_handler, event_loop):
    exception = errors.UpsiesError('my exception')
    callback = Mock(side_effect=exception)
    coro = AsyncMock(return_value='foo')()

    task = utils.run_task(coro, callback=callback)
    return_value = await task
    assert return_value == 'foo'

    assert callback.call_args_list == [call('foo')]

    assert_unhandled_exception(unhandled_exception_handler, exception)

@pytest.mark.parametrize(
    argnames='args, kwargs',
    argvalues=(
        ((), {}),
        (('foo', 'bar'), {}),
        ((), {'baz': 123}),
        (('foo', 'bar'), {'baz': 123}),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_run_async(args, kwargs, event_loop):
    function = Mock()
    return_value = await utils.run_async(function, *args, **kwargs)
    assert return_value == function.return_value
    assert function.call_args_list == [call(*args, **kwargs)]
